#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module containing classes to read labeled datasets.
Allows to read labeled datasets from exports or from the cloud.

Classes
-------
    LabeledDatasetReader
        The LabeledDatasetReader class to read labeled dataset.
    ExportReader
        The ExportReader class to read labeled dataset from the export.
    DLSDatasetReader
        The DLSDatasetReader class to read labeled dataset from the cloud.

Examples
--------
    >>> from ads.common import auth as authutil
    >>> from ads.data_labeling import LabeledDatasetReader
    >>> ds_reader = LabeledDatasetReader.from_export(
    ...    path="oci://bucket_name@namespace/dataset_metadata.jsonl",
    ...    auth=authutil.api_keys(),
    ...    materialize=True
    ... )
    >>> ds_reader.info()
        ------------------------------------------------------------------------
        annotation_type	                                            SINGLE_LABEL
        compartment_id	                                        TEST_COMPARTMENT
        dataset_id                                                  TEST_DATASET
        dataset_name                                           test_dataset_name
        dataset_type                                                        TEXT
        labels	                                                   ['yes', 'no']
        records_path                                             path/to/records
        source_path                                              path/to/dataset

    >>> ds_reader.read()
                                 Path            Content            Annotations
        -----------------------------------------------------------------------
        0   path/to/the/content/file1       file content                    yes
        1   path/to/the/content/file2       file content                     no
        2   path/to/the/content/file3       file content                     no

    >>> next(ds_reader.read(iterator=True))
        ("path/to/the/content/file1", "file content", "yes")

    >>> next(ds_reader.read(iterator=True, chunksize=2))
        [("path/to/the/content/file1", "file content", "yes"),
        ("path/to/the/content/file2", "file content", "no")]

    >>> next(ds_reader.read(chunksize=2))
                                Path            Content            Annotations
        ----------------------------------------------------------------------
        0   path/to/the/content/file1       file content                    yes
        1   path/to/the/content/file2       file content                     no

    >>> ds_reader = LabeledDatasetReader.from_DLS(
    ...    dataset_id="dataset_OCID",
    ...    compartment_id="compartment_OCID",
    ...    auth=authutil.api_keys(),
    ...    materialize=True
    ... )
"""

from functools import lru_cache
from typing import Any, Dict, Generator, Tuple, Union

import pandas as pd
from ads.common import auth as authutil
from ads.common.serializer import Serializable
from ads.data_labeling.interface.reader import Reader
from ads.data_labeling.reader.metadata_reader import Metadata, MetadataReader
from ads.data_labeling.reader.record_reader import RecordReader
from ads.data_labeling.constants import (
    FORMATS_TO_ANNOTATION_TYPE,
    ANNOTATION_TYPE_TO_FORMATS,
)
from ads.config import NB_SESSION_COMPARTMENT_OCID, JOB_RUN_COMPARTMENT_OCID

_LABELED_DF_COLUMNS = ["Path", "Content", "Annotations"]


class LabeledDatasetReader:
    """The labeled dataset reader class.

    Methods
    -------
    info(self) -> Metadata
        Gets labeled dataset metadata.
    read(self, iterator: bool = False) -> Union[Generator[Any, Any, Any], pd.DataFrame]
        Reads labeled dataset.
    from_export(cls, path: str, auth: Dict = None, encoding="utf-8", materialize: bool = False) -> "LabeledDatasetReader"
        Constructs a Labeled Dataset Reader instance.

    Examples
    --------
    >>> from ads.common import auth as authutil
    >>> from ads.data_labeling import LabeledDatasetReader

    >>> ds_reader = LabeledDatasetReader.from_export(
    ...    path="oci://bucket_name@namespace/dataset_metadata.jsonl",
    ...    auth=authutil.api_keys(),
    ...    materialize=True
    ... )

    >>> ds_reader = LabeledDatasetReader.from_DLS(
    ...    dataset_id="dataset_OCID",
    ...    compartment_id="compartment_OCID",
    ...    auth=authutil.api_keys(),
    ...    materialize=True
    ... )

    >>> ds_reader.info()
        ------------------------------------------------------------------------
        annotation_type	                                            SINGLE_LABEL
        compartment_id	                                        TEST_COMPARTMENT
        dataset_id                                                  TEST_DATASET
        dataset_name                                           test_dataset_name
        dataset_type                                                        TEXT
        labels	                                                   ['yes', 'no']
        records_path                                             path/to/records
        source_path                                              path/to/dataset

    >>> ds_reader.read()
                                 Path            Content            Annotations
        -----------------------------------------------------------------------
        0   path/to/the/content/file1       file content                    yes
        1   path/to/the/content/file2       file content                     no
        2   path/to/the/content/file3       file content                     no

    >>> next(ds_reader.read(iterator=True))
        ("path/to/the/content/file1", "file content", "yes")

    >>> next(ds_reader.read(iterator=True, chunksize=2))
        [("path/to/the/content/file1", "file content", "yes"),
        ("path/to/the/content/file2", "file content", "no")]

    >>> next(ds_reader.read(chunksize=2))
                                Path            Content            Annotations
        ----------------------------------------------------------------------
        0   path/to/the/content/file1       file content                    yes
        1   path/to/the/content/file2       file content                     no
    """

    def __init__(self, reader: Reader):
        """Initializes the labeled dataset reader instance.

        Parameters
        ----------
        reader: Reader
            The Reader instance which reads and extracts the labeled dataset.
        """
        self._reader = reader

    @classmethod
    def from_export(
        cls,
        path: str,
        auth: dict = None,
        encoding: str = "utf-8",
        materialize: bool = False,
        include_unlabeled: bool = False,
    ) -> "LabeledDatasetReader":
        """Constructs Labeled Dataset Reader instance.

        Parameters
        ----------
        path: str
            The metadata file path, can be either local or object storage path.
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        encoding: (str, optional). Defaults to 'utf-8'.
            Encoding for files.
        materialize: (bool, optional). Defaults to False.
            Whether the content of the dataset file should be loaded or it should return the file path to the content.
            By default the content will not be loaded.

        Returns
        -------
        LabeledDatasetReader
            The LabeledDatasetReader instance.
        """
        auth = auth or authutil.default_signer()

        return cls(
            reader=ExportReader(
                path=path,
                auth=auth,
                encoding=encoding,
                materialize=materialize,
                include_unlabeled=include_unlabeled,
            )
        )

    @classmethod
    def from_DLS(
        cls,
        dataset_id: str,
        compartment_id: str = None,
        auth: dict = None,
        encoding: str = "utf-8",
        materialize: bool = False,
        include_unlabeled: bool = False,
    ) -> "LabeledDatasetReader":
        """Constructs Labeled Dataset Reader instance.

        Parameters
        ----------
        dataset_id: str
            The dataset OCID.
        compartment_id: str. Defaults to the compartment_id from the env variable.
            The compartment OCID of the dataset.
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        encoding: (str, optional). Defaults to 'utf-8'.
            Encoding for files.
        materialize: (bool, optional). Defaults to False.
            Whether the content of the dataset file should be loaded or it should return the file path to the content.
            By default the content will not be loaded.

        Returns
        -------
        LabeledDatasetReader
            The LabeledDatasetReader instance.
        """
        if compartment_id is None:
            compartment_id = NB_SESSION_COMPARTMENT_OCID or JOB_RUN_COMPARTMENT_OCID

        if not compartment_id:
            raise ValueError("The `compartment_id` must be provided.")

        return cls(
            reader=DLSDatasetReader(
                compartment_id=compartment_id,
                dataset_id=dataset_id,
                auth=auth or authutil.default_signer(),
                encoding=encoding,
                materialize=materialize,
                include_unlabeled=include_unlabeled,
            )
        )

    def info(self) -> Serializable:
        """Gets the labeled dataset metadata.

        Returns
        -------
        Metadata
            The labeled dataset metadata.
        """
        return self._reader.info()

    def read(
        self, iterator: bool = False, format: str = None, chunksize: int = None
    ) -> Union[Generator[Any, Any, Any], pd.DataFrame]:
        """Reads the labeled dataset records.

        Parameters
        ----------
        iterator: (bool, optional). Defaults to False.
            True if the result should be represented as a Generator.
            Fasle if the result should be represented as a Pandas DataFrame.
        format: (str, optional). Defaults to None.
            Output format of annotations. Can be None, "spacy" or "yolo".
        chunksize: (int, optional). Defaults to None.
            The number of records that should be read in one iteration.
            The result will be returned in a generator format.

        Returns
        -------
        Union[
            Generator[Tuple[str, str, Any], Any, Any],
            Generator[List[Tuple[str, str, Any]], Any, Any],
            Generator[pd.DataFrame, Any, Any],
            pd.DataFrame
        ]
            `pd.Dataframe` if `iterator` and `chunksize` are not specified.
            `Generator[pd.Dataframe] ` if `iterator` equal to False and `chunksize` is specified.
            `Generator[List[Tuple[str, str, Any]]]` if `iterator` equal to True and `chunksize` is specified.
            `Generator[Tuple[str, str, Any]]` if `iterator` equal to True and `chunksize` is not specified.
        """

        if chunksize:
            return self._bulk_read(
                iterator=iterator, format=format, chunksize=chunksize
            )

        if iterator:
            return self._reader.read(format=format)

        return pd.DataFrame(
            self._reader.read(format=format), columns=_LABELED_DF_COLUMNS
        )

    def _bulk_read(
        self, iterator: bool = False, format: str = None, chunksize: int = None
    ) -> Generator[Union[pd.DataFrame, Tuple[str, str, Any]], Any, Any]:
        """Reads the labeled dataset records by chunks.

        Parameters
        ----------
        iterator: (bool, optional). Defaults to False.
            True if the result should be represented as a Generator.
            Fasle if the result should be represented as a Pandas DataFrame.
        format: (str, optional). Defaults to None.
            Output format of annotations. Can be None, "spacy" or "yolo".
        chunksize: (int, optional). Defaults to None.
            The number of records that should be read in one iteration.
            Result will be represented as a generator.

        Yields
        -------
        Generator[Union[pd.DataFrame, Tuple[str, str, Any]], Any, Any]
            The generator that yields records either in Dataframe format or Tuple.

        Raises
        ------
            ValueError: If chunksize is empty or not a positive integer.
        """
        if not chunksize or not isinstance(chunksize, int) or chunksize < 1:
            raise ValueError("`chunksize` must be a positive integer.")

        result = []
        i = 0
        for record in self._reader.read(format=format):
            result.append(record)
            i += 1
            if i >= chunksize:
                yield result if iterator else pd.DataFrame(
                    result, columns=_LABELED_DF_COLUMNS
                )
                result = []
                i = 0
        if result:
            yield result if iterator else pd.DataFrame(
                result, columns=_LABELED_DF_COLUMNS
            )


class DLSDatasetReader(Reader):
    """The DLSDatasetReader class to read labeled dataset from the cloud.

    Methods
    -------
    info(self) -> Metadata
        Gets the labeled dataset metadata.
    read(self) -> Generator[Tuple, Any, Any]
        Reads the labeled dataset.
    """

    def __init__(
        self,
        dataset_id: str,
        compartment_id: str,
        auth: Dict,
        encoding="utf-8",
        materialize: bool = False,
        include_unlabeled: bool = False,
    ):
        """Initializes the DLS dataset reader instance.

        Parameters
        ----------
        dataset_id: str
            The dataset OCID.
        compartment_id: str
            The compartment OCID of the dataset.
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        encoding: (str, optional). Defaults to 'utf-8'.
            Encoding for files. The encoding is used to extract the metadata information
            of the labeled dataset and also to extract the content of the text dataset records.
        materialize: (bool, optional). Defaults to False.
            Whether the content of dataset files should be loaded/materialized or not.
            By default the content will not be materialized.
        include_unlabeled: (bool, optional). Defaults to False.
            Whether to load the unlabeled records or not.

        Raises
        ------
            ValueError: When dataset_id is empty or not a string.
            TypeError: When dataset_id not a string.
        """
        if not dataset_id:
            raise ValueError("The dataset OCID must be specified.")

        if not isinstance(dataset_id, str):
            raise TypeError("The dataset_id must be a string.")

        if not compartment_id:
            raise ValueError("The compartment OCID must be specified.")

        self.dataset_id = dataset_id
        self.compartment_id = compartment_id
        self.auth = auth or authutil.default_signer()
        self.encoding = encoding
        self.materialize = materialize
        self.include_unlabeled = include_unlabeled

    @lru_cache(maxsize=1)
    def info(self) -> Metadata:
        """Gets the labeled dataset metadata.

        Returns
        -------
        Metadata
            The labeled dataset metadata.
        """
        return MetadataReader.from_DLS(
            compartment_id=self.compartment_id,
            dataset_id=self.dataset_id,
        ).read()

    def read(self, format: str = None) -> Generator[Tuple, Any, Any]:
        """Reads the labeled dataset records.

        Parameters
        ----------
        format: (str, optional). Defaults to None.
            Output format of annotations. Can be None, "spacy" for dataset
            Entity Extraction type or "yolo" for Object Detection type.
            When None, it outputs List[NERItem] or List[BoundingBoxItem].
            When "spacy", it outputs List[Tuple].
            When "yolo", it outputs List[List[Tuple]].

        Returns
        -------
        Generator[Tuple, Any, Any]
            The labeled dataset records.
        """

        metadata = self.info()

        records_reader = RecordReader.from_DLS(
            dataset_type=metadata.dataset_type,
            annotation_type=metadata.annotation_type,
            dataset_source_path=metadata.source_path,
            compartment_id=self.compartment_id,
            dataset_id=self.dataset_id,
            auth=self.auth,
            encoding=self.encoding,
            materialize=self.materialize,
            include_unlabeled=self.include_unlabeled,
            format=format,
            categories=metadata.labels,
        )
        return records_reader.read()


class ExportReader(Reader):
    """The ExportReader class to read labeled dataset from the export.

    Methods
    -------
    info(self) -> Metadata
        Gets the labeled dataset metadata.
    read(self) -> Generator[Tuple, Any, Any]
        Reads the labeled dataset.
    """

    def __init__(
        self,
        path: str,
        auth: Dict = None,
        encoding="utf-8",
        materialize: bool = False,
        include_unlabeled: bool = False,
    ):
        """Initializes the labeled dataset export reader instance.

        Parameters
        ----------
        path: str
            The metadata file path, can be either local or object storage path.
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        encoding: (str, optional). Defaults to 'utf-8'.
            Encoding for files. The encoding is used to extract the metadata information
            of the labeled dataset and also to extract the content of the text dataset records.
        materialize: (bool, optional). Defaults to False.
            Whether the content of dataset files should be loaded/materialized or not.
            By default the content will not be materialized.
        include_unlabeled: (bool, optional). Defaults to False.
            Whether to load the unlabeled records or not.

        Raises
        ------
            ValueError: When path is empty or not a string.
            TypeError: When path not a string.
        """

        if not path:
            raise ValueError("The parameter `path` is required.")

        if not isinstance(path, str):
            raise TypeError("The parameter `path` must be a string.")

        self.path = path
        self.auth = auth or authutil.default_signer()
        self.encoding = encoding
        self.materialize = materialize
        self.include_unlabeled = include_unlabeled

    @lru_cache(maxsize=1)
    def info(self) -> Metadata:
        """Gets the labeled dataset metadata.

        Returns
        -------
        Metadata
            The labeled dataset metadata.
        """
        return MetadataReader.from_export_file(
            path=self.path,
            auth=self.auth,
        ).read()

    def read(self, format: str = None) -> Generator[Tuple, Any, Any]:
        """Reads the labeled dataset records.

        Parameters
        ----------
        format: (str, optional). Defaults to None.
            Output format of annotations. Can be None, "spacy" for dataset
            Entity Extraction type or "yolo" for Object Detection type.
            When None, it outputs List[NERItem] or List[BoundingBoxItem].
            When "spacy", it outputs List[Tuple].
            When "yolo", it outputs List[List[Tuple]].

        Returns
        -------
        Generator[Tuple, Any, Any]
            The labeled dataset records.
        """
        metadata = self.info()
        if (
            format
            and isinstance(format, str)
            and (
                format.lower() not in FORMATS_TO_ANNOTATION_TYPE.keys()
                or FORMATS_TO_ANNOTATION_TYPE[format.lower()]
                != metadata.annotation_type
            )
        ):
            raise TypeError(
                "Wrong format. `format` can only be None or "
                f"`{ANNOTATION_TYPE_TO_FORMATS[metadata.annotation_type]}`."
            )

        records_reader = RecordReader.from_export_file(
            path=metadata.records_path or self.path,
            dataset_type=metadata.dataset_type,
            annotation_type=metadata.annotation_type,
            dataset_source_path=metadata.source_path,
            auth=self.auth,
            encoding=self.encoding,
            materialize=self.materialize,
            include_unlabeled=self.include_unlabeled,
            format=format,
            categories=metadata.labels,
            includes_metadata=not metadata.records_path,
        )
        return records_reader.read()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents an Unknown feature type.

Classes:
    Text
        The Unknown feature type.
"""
from ads.feature_engineering.feature_type.base import FeatureType


class Unknown(FeatureType):
    """
    Type representing third-party dtypes.

    Attributes
    ----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.
    """

    description = "Type representing unknown type."

    @classmethod
    def feature_domain(cls):
        """
        Returns
        -------
        None
            Nothing.
        """
        return None



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import ads
from ads.common import oci_client as oc
import fsspec
import uuid
import os
from ads.jobs import Job
import oci
from collections import defaultdict
import ipaddress
from ads.opctl.distributed.common.abstract_cluster_provider import ClusterProvider
from ads.opctl.diagnostics.requirement_exception import RequirementException
import urllib.parse
import requests

cp = ClusterProvider(mode="MAIN")
auth = cp.get_oci_auth()
cf = oc.OCIClientFactory(**auth)


class PortsChecker:
    def __init__(self, subnet_id=None, auth=None, **kwargs):
        if subnet_id:
            self.subnet_id = subnet_id
        else:
            job_ocid = os.environ["JOB_OCID"]
            jobDef = Job.from_datascience_job(job_ocid)
            self.subnet_id = jobDef.infrastructure.subnet_id
        self.authinfo = auth or ads.common.auth.default_signer()

    def _fetch_security_list(self):
        core_client = oci.core.VirtualNetworkClient(**self.authinfo)
        subnet = core_client.get_subnet(self.subnet_id).data
        cidr = subnet.cidr_block
        security_list_ids = subnet.security_list_ids
        print(f"{cidr=}")
        # print(type(security_list_ids))
        for id in security_list_ids:
            print(f"{id=}")

        port_open_details = defaultdict(set)
        for id in security_list_ids:
            security_list_info = core_client.get_security_list(id).data
            # print(f"Egress: {security_list_info.egress_security_rules}")
            ingress_security_rules = security_list_info.ingress_security_rules
            for rule in ingress_security_rules:
                tcp_options = (
                    rule.tcp_options.destination_port_range if rule.tcp_options else {}
                )
                source = rule.source
                if tcp_options:
                    port_open_details[source].add((tcp_options.min, tcp_options.max))
                if (
                    rule.protocol == "all" or rule.protocol.lower() == "6"
                ) and rule.tcp_options is None:
                    port_open_details[source].add((0,))

        for key in port_open_details:
            port_open_details[key] = sorted(port_open_details[key])
        return port_open_details

    def _flatten_candidate_port_range(self, port_list):
        """
        Ports are opened by providing the CIDR block of the source host. We need to check only within the hosts that intersects the IP of the job instance
        """
        self_ip = cp.find_self_ip(self.authinfo)
        flattened_port_list = []
        for cidr in port_list:
            if ipaddress.ip_address(self_ip) in ipaddress.ip_network(cidr):
                flattened_port_list += port_list[cidr]
        return sorted(flattened_port_list)

    def check_ports(self, ports, *kwargs):
        port_list = self._flatten_candidate_port_range(self._fetch_security_list())
        unsatisfied_ports = []
        print(f"Checking if ports are open", flush=True)
        print(f"Open port information: {port_list}", flush=True)
        if port_list[0] == (0,):
            print(f"Subnet has all ports open.")
            return True
        for port in ports:
            print(f"Checking if port {port} is open", flush=True)
            if str(port).lower() == "all":
                if port_list[0] != (0,):
                    print(
                        f"Requires ALL ports for this subnet to be open. Open ports are: {'None' if len(port_list) == 0 else port_list}"
                    )
                    unsatisfied_ports.append(port)
                    break
                else:
                    return True
            elif port in port_list:
                print(f"found port in open port list: {port_list}", flush=True)
                continue
            else:
                if len(port.split("-")) == 1:
                    found = False
                    for port_range in port_list:
                        if int(port) >= port_range[0] and int(port) <= port_range[1]:
                            print(
                                f"found port: {port} in range {port_range}", flush=True
                            )
                            found = True
                            break
                    if not found:
                        print(
                            f"Not found port: {port} in {port_list}!!! Check security list",
                            flush=True,
                        )
                        unsatisfied_ports.append(port)
                else:
                    minport = int(port.split("-")[0])
                    maxport = int(port.split("-")[1])
                    found = False
                    curr_minport = minport
                    # It might be better to normalize the range in port_list
                    # to remove any possible overlaps between the ranges. That
                    # will make search much easier than below.
                    for port_range in port_list:
                        if curr_minport >= port_range[1]:
                            continue

                        if curr_minport >= port_range[0]:
                            if maxport <= port_range[1]:
                                print(f"found port open in range {port_list}")
                                found = True
                                break
                            else:
                                curr_minport = (
                                    port_range[1] + 1
                                )  # Can fail if the next range starts at maxport or less than maxport
                    if not found:
                        print(
                            f"Not found port: {port} in {port_list}!!! Check security list",
                            flush=True,
                        )
                        unsatisfied_ports.append(port)
        if len(unsatisfied_ports) == 0:
            return True
        else:
            raise RequirementException(
                f"Following ports are not opened: {unsatisfied_ports}. Update security list corresponding to the subnet: {self.subnet_id}",
            )


class Default:
    def __init__(self, auth, *kwargs):
        self.auth = auth

    def job_info_access_policy(self, **kwargs):
        print(f"Checking Job info access")
        try:
            job_ocid = os.environ["JOB_OCID"]
            Job.from_datascience_job(job_ocid)
            print(f"Passed Job info access")
        except oci.exceptions.ServiceError as se:
            print(se)
            raise RequirementException(
                f"Could not access Job Definition details. To fix this: \n \
                Allow dynamic-group <distributed_training_job_runs> to use data-science-family in compartment <your_compartment_name>"
            )

    def job_vcn_info_access_policy(self, **kwargs):
        print(f"Checking VCN info access")

        job_ocid = os.environ["JOB_OCID"]
        jobDef = Job.from_datascience_job(job_ocid)

        subnet_id = jobDef.infrastructure.subnet_id
        client = cf.create_client(oci.core.VirtualNetworkClient)
        try:
            client.get_subnet(subnet_id).data
            print(f"Passed VCN info access")
        except oci.exceptions.ServiceError as se:
            print(se)
            raise RequirementException(
                f"Could not access Job Definition details. To fix this: \n \
                Allow dynamic-group <distributed_training_job_runs> to use virtual-network-family in compartment <your_compartment_name>"
            )

    def work_dir_access(self):
        test_file_name = f"{os.environ['OCI__WORK_DIR']}/{str(uuid.uuid4())}"
        bucket_name = urllib.parse.urlparse(test_file_name).username
        print(f"Checking Object storage access to {os.environ['OCI__WORK_DIR']}")
        try:
            with fsspec.open(test_file_name, "w", **self.auth) as tf:
                tf.write("sample")
            print(f"Passed object storage access")
        except oci.exceptions.ServiceError as se:
            print(se)
            raise RequirementException(
                "Could not access Job Definition details. To fix this: \n \
                Allow dynamic-group <distributed_training_job_runs> to manage objects in compartment your_compartment_name where all {target.bucket.name="
                + bucket_name
                + "}"
            )

    def check_internet(self, **kwargs):
        print(f"Checking for internet")
        try:
            res = requests.get("https://www.oracle.com", timeout=5)
        except requests.exceptions.ConnectionError as ce:
            print(ce)
            raise RequirementException(
                f"Could not access internet. To fix this: \n \
                Add NAT gateway to your subnet. If you are behind proxy, add https_proxy and http_proxy to the environment variables"
            )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from getopt import gnu_getopt
import shlex
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.pipeline import Pipeline, PipelineRun, LogType

MAXIMUM_TIMEOUT_SECONDS = 1800
LOG_RECORDS_LIMIT = 100


def pipeline(line, cell=None):
    opts, args = gnu_getopt(
        shlex.split(line),
        "f:o:l:m:n:wxjpstdh",
        longopts=[
            "file=",
            "ocid=",
            "log-type=",
            "max-wait-seconds=",
            "number=",
            "watch",
            "text",
            "no-delete-related-job-runs",
            "no-delete-related-pipeline-runs",
            "succeed-on-not-found",
            "tail",
            "head",
        ],
    )

    PIPELINE_MAGIC_COMMANDS = {
        "run": pipeline_run,
        "log": pipeline_log,
        "cancel": pipeline_cancel,
        "delete": pipeline_delete,
        "show": pipeline_show,
        "status": pipeline_status,
    }

    opts_dict = {k: v for k, v in opts}
    if len(args) == 0 and "-h" in opts_dict:
        print(
            f"""
Usage: pipeline [SUBCOMMAND]
Subcommand:
    run, run a pipeline from YAML or an existing ocid.
    log, stream the logs from pipeline run.
    cancel, cancel a pipeline run.
    delete, delete pipeline or pipeline run.
    show, show the pipeline orchestration.
    status, show the real-time status of a pipeline run.

Run pipeline [SUBCOMMAND] -h to see more details.
        """
        )
        return
    if args[0] not in PIPELINE_MAGIC_COMMANDS:
        commands = ", ".join(list(PIPELINE_MAGIC_COMMANDS.keys()))
        print(f"`pipeline` expects subcommand {commands}.")
        return

    PIPELINE_MAGIC_COMMANDS[args[0]](opts_dict, args)


def pipeline_run(options, args):
    if "-h" in options:
        print(
            f"""
Usage: pipeline run [OPTIONS]
Options:
    -f, --file, optional, uri to the YAML.
    -o, --ocid, optional, ocid of existing pipeline.
    -w, --watch, optional, a flag indicating that pipeline run will be watched after submission.
    -l, --log-type, optional, should be either custom_log or service_log. default is "custom_log".
    -h, show this help message.
        """
        )
        return
    file = None
    ocid = None
    if "-f" in options:
        file = options["-f"]
    elif "--file" in options:
        file = options["--file"]
    if "-o" in options:
        ocid = options["-o"]
    elif "--ocid" in options:
        ocid = options["--ocid"]

    if ocid:
        pipeline = Pipeline.from_id(ocid)
    elif file:
        pipeline = Pipeline.from_yaml(uri=file)
        pipeline.create()
    else:
        print(
            f"`pipeline run` expects YAML file uri or pipeline ocid. Use `pipeline run` with options."
        )
        return

    print("Pipeline OCID:", pipeline.id)
    pipeline_run = pipeline.run()
    print("Pipeline Run OCID:", pipeline_run.id)
    print(pipeline)

    if "-w" in options or "--watch" in options:
        log_type = (
            options.get("-l", None)
            or options.get("--log-type", None)
            or LogType.CUSTOM_LOG
        )
        if log_type not in (LogType.CUSTOM_LOG, LogType.SERVICE_LOG):
            print(
                "Log type should be either custom_log or service_log. Using default custom_log."
            )
            log_type = LogType.CUSTOM_LOG
        print("Streaming logs...")
        pipeline_run.watch(log_type=log_type)


def pipeline_log(options, args):
    if "-h" in options:
        print(
            f"""
Usage: pipeline log [OPTIONS] [RUN_ID]
Options:
    -l, --log-type, optional, should be either custom_log, service_log or None. default is None.
    -t, --tail, a flag to show the most recent log records.
    -d, --head, a flag to show the preceding log records.
    -n, --number, number of lines of logs to be printed. Defaults to 100.
    -h, show this help message.
        """
        )
        return
    if len(args) < 2:
        raise ValueError("Pipeline Run ID must be provided.")
    log_type = options.get("-l", None) or options.get("--log-type", None)
    if log_type and log_type not in (LogType.CUSTOM_LOG, LogType.SERVICE_LOG):
        print("Log type should be either custom_log, service_log or None.")
        return

    limit = LOG_RECORDS_LIMIT
    if "-n" in options:
        limit = int(options["-n"])
    elif "--number" in options:
        limit = int(options["--number"])

    pipeline_run = PipelineRun.from_ocid(args[1])
    if "-t" in options or "--tail" in options:
        pipeline_run.logs(log_type=log_type).tail(limit=limit)
    elif "-d" in options or "--head" in options:
        pipeline_run.logs(log_type=log_type).head(limit=limit)
    else:
        pipeline_run.watch(log_type=log_type)


def pipeline_cancel(options, args):
    if "-h" in options:
        print(
            f"""
Usage: pipeline cancel [RUN_ID]
Options:
    -h, show this help message.
        """
        )
        return
    if len(args) < 2:
        raise ValueError("Pipeline Run ID must be provided.")

    pipeline_run = PipelineRun.from_ocid(args[1])
    pipeline_run.cancel()
    print(f"Pipeline Run {args[1]} has been cancelled.")


def pipeline_delete(options, args):
    if "-h" in options:
        print(
            f"""
Usage: pipeline delete [OCID]
Options:
    -j, --no-delete-related-job-runs, a flag to not delete the related job runs.
    -p, --no-delete-related-pipeline-runs, a flag to not delete related pipeline runs.
    -m, --max-wait-seconds, integer, maximum wait time in second for delete to complete. Defaults to 1800.
    -s, --succeeded-on-not-found, to flag to return successfully if the data we're waiting on is not found.
    -h, show this help message.
        """
        )
        return
    if len(args) < 2:
        raise ValueError("Pipeline or Pipeline Run ID must be provided.")

    delete_related_job_runs = True
    delete_related_pipeline_runs = True
    max_wait_seconds = MAXIMUM_TIMEOUT_SECONDS
    succeed_on_not_found = False

    if "-j" in options or "--no-delete-related-job-runs" in options:
        delete_related_job_runs = False

    if "-p" in options or "--no-delete-related-pipeline-runs" in options:
        delete_related_pipeline_runs = False

    if "-m" in options:
        max_wait_seconds = int(options["-m"])
    elif "--max-wait-seconds" in options:
        max_wait_seconds = int(options["--max-wait-seconds"])

    if "-s" in options or "--succeeded-on-not-found" in options:
        succeed_on_not_found = True

    id = args[1]
    if "datasciencepipelinerun" in id:
        pipeline_run = PipelineRun.from_ocid(id)
        pipeline_run.delete(
            delete_related_job_runs=delete_related_job_runs,
            max_wait_seconds=max_wait_seconds,
        )
        print(f"Pipeline Run {args[1]} has been deleted.")
    elif "datasciencepipeline" in id:
        pipeline = Pipeline.from_ocid(id)
        pipeline.delete(
            delete_related_pipeline_runs=delete_related_pipeline_runs,
            delete_related_job_runs=delete_related_job_runs,
            max_wait_seconds=max_wait_seconds,
            succeed_on_not_found=succeed_on_not_found,
        )
        print(f"Pipeline {args[1]} has been deleted.")
    else:
        print(f"`pipeline delete` expects a valid pipeline or pipeline run id.")


def pipeline_show(options, args):
    if "-h" in options:
        print(
            f"""
Usage: pipeline show [OCID]
Options:
    -h, show this help message.
        """
        )
        return
    if len(args) < 2:
        raise ValueError("Pipeline ID must be provided.")

    pipeline = Pipeline.from_ocid(args[1])
    pipeline.show()


def pipeline_status(options, args):
    if "-h" in options:
        print(
            f"""
Usage: pipeline status [OPTIONS] [RUN_ID]
Options:
    -x, --text, optional, a flag to show the status in text format.
    -w, --watch, optional, a flag to wait until the completion of the pipeline run.
    If set, the rendered graph will be updated until the completion of the pipeline run,
    otherwise will render one graph with the current status.
    -h, show this help message.
        """
        )
        return
    if len(args) < 2:
        raise ValueError("Pipeline Run ID must be provided.")

    pipeline_run = PipelineRun.from_ocid(args[1])

    if "-x" in options or "--text" in options:
        if "-w" in options or "--watch" in options:
            pipeline_run.show(mode="text")
        else:
            pipeline_run.show(mode="text")
    else:
        if "-w" in options or "--watch" in options:
            pipeline_run.show(wait=True)
        else:
            pipeline_run.show()


@runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
def load_ipython_extension(ipython):
    ipython.register_magic_function(pipeline)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that provides the decorator helping to add runtime dependencies in functions.

Examples
--------
>>> @runtime_dependency(module="pandas", short_name="pd")
... def test_function()
...     print(pd)

>>> @runtime_dependency(module="pandas", object="DataFrame", short_name="df")
... def test_function()
...     print(df)

>>> @runtime_dependency(module="pandas", short_name="pd")
... @runtime_dependency(module="pandas", object="DataFrame", short_name="df")
... def test_function()
...     print(df)
...     print(pd)

>>> @runtime_dependency(module="pandas", object="DataFrame", short_name="df", install_from="ads[optional]")
... def test_function()
...     pass

>>> @runtime_dependency(module="pandas", object="DataFrame", short_name="df", err_msg="Custom error message.")
... def test_function()
...     pass
"""

import importlib
import logging
from functools import wraps
from typing import Any, Callable

from .utils import _get_original_func

logger = logging.getLogger(__name__)

MODULE_NOT_FOUND_ERROR = (
    "The `{module}` module was not found. Please run `pip install {package}`."
)
IMPORT_ERROR = "Cannot import name `{object}` from `{module}`."


class OptionalDependency:
    LABS = "oracle-ads[labs]"
    BOOSTED = "oracle-ads[boosted]"
    NOTEBOOK = "oracle-ads[notebook]"
    TEXT = "oracle-ads[text]"
    VIZ = "oracle-ads[viz]"
    DATA = "oracle-ads[data]"
    OPCTL = "oracle-ads[opctl]"
    MYSQL = "oracle-ads[mysql]"
    BDS = "oracle-ads[bds]"
    PYTORCH = "oracle-ads[torch]"
    TENSORFLOW = "oracle-ads[tensorflow]"
    GEO = "oracle-ads[geo]"
    ONNX = "oracle-ads[onnx]"
    OPTUNA = "oracle-ads[optuna]"
    SPARK = "oracle-ads[spark]"
    HUGGINGFACE = "oracle-ads[huggingface]"


def runtime_dependency(
    module: str,
    short_name: str = "",
    object: str = None,
    install_from: str = None,
    err_msg: str = "",
    is_for_notebook_only=False,
):
    """The decorator which is helping to add runtime dependencies to functions.

    Parameters
    ----------
    module: str
        The module name to be imported.
    short_name: (str, optional). Defaults to empty string.
        The short name for the imported module.
    object: (str, optional). Defaults to None.
        The name of the object to be imported. Can be a function or a class, or
        any variable provided by module.
    install_from: (str, optional). Defaults to None.
        The parameter helping to answer from where the required dependency can be installed.
    err_msg: (str, optional). Defaults to empty string.
        The custom error message.
    is_for_notebook_only: (bool, optional). Defaults to False.
        If the value of this flag is set to True, the dependency will be added only
        in case when the current environment is a jupyter notebook.

    Raises
    ------
    ModuleNotFoundError
        In case if requested module not found.
    ImportError
        In case if object cannot be imported from the module.

    Examples
    --------
    >>> @runtime_dependency(module="pandas", short_name="pd")
    ... def test_function()
    ...     print(pd)

    >>> @runtime_dependency(module="pandas", object="DataFrame", short_name="df")
    ... def test_function()
    ...     print(df)

    >>> @runtime_dependency(module="pandas", short_name="pd")
    ... @runtime_dependency(module="pandas", object="DataFrame", short_name="df")
    ... def test_function()
    ...     print(df)
    ...     print(pd)

    >>> @runtime_dependency(module="pandas", object="DataFrame", short_name="df", install_from="ads[optional]")
    ... def test_function()
    ...     pass

    >>> @runtime_dependency(module="pandas", object="DataFrame", short_name="df", err_msg="Custom error message.")
    ... def test_function()
    ...     pass
    """

    def decorator(func: Callable):
        @wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            from ads.common.utils import is_notebook

            if not is_for_notebook_only or is_notebook():
                assert module, "The parameter `module` must be provided."
                assert isinstance(
                    module, str
                ), "The parameter `module` must be a string."
                try:
                    _package_name = module.split(".")[0]
                    # importing module
                    _module = importlib.import_module(module)
                    # checking if object parameter is specified then inporting
                    # the object from the module. The object can be a function
                    # or a class or any module vatiable.
                    _module = getattr(_module, object) if object else _module
                    # retrieving the original function from the decorated one.
                    # This is necessary when the chain of decorators is used.
                    # The only original function should be injected with the
                    # dependency.
                    original_func = _get_original_func(func)
                    _module_name = module.split(".")[-1]
                    # Injecting the imported module into global variables of the
                    # original function.
                    original_func.__globals__[
                        short_name or object or _module_name
                    ] = _module
                except ModuleNotFoundError:
                    raise ModuleNotFoundError(
                        err_msg
                        or MODULE_NOT_FOUND_ERROR.format(
                            module=module, package=install_from or _package_name
                        )
                    )
                except AttributeError:
                    raise ImportError(
                        err_msg or IMPORT_ERROR.format(object=object, module=module)
                    )
            return func(*args, **kwargs)

        return wrapper

    return decorator



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from ads.common import logger
from ads.model.extractor.pytorch_extractor import PytorchExtractor
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.model.generic_model import FrameworkSpecificModel
from ads.model.model_properties import ModelProperties
from ads.model.serde.model_serializer import PyTorchModelSerializerType
from ads.model.common.utils import (
    DEPRECATE_AS_ONNX_WARNING,
    DEPRECATE_USE_TORCH_SCRIPT_WARNING,
)
from ads.model.serde.common import SERDE

ONNX_MODEL_FILE_NAME = "model.onnx"
PYTORCH_MODEL_FILE_NAME = "model.pt"


class PyTorchModel(FrameworkSpecificModel):
    """PyTorchModel class for estimators from Pytorch framework.

    Attributes
    ----------
    algorithm: str
        The algorithm of the model.
    artifact_dir: str
        Artifact directory to store the files needed for deployment.
    auth: Dict
        Default authentication is set using the `ads.set_auth` API. To override the
        default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create
        an authentication signer to instantiate an IdentityClient object.
    estimator: Callable
        A trained pytorch estimator/model using Pytorch.
    framework: str
        "pytorch", the framework name of the model.
    hyperparameter: dict
        The hyperparameters of the estimator.
    metadata_custom: ModelCustomMetadata
        The model custom metadata.
    metadata_provenance: ModelProvenanceMetadata
        The model provenance metadata.
    metadata_taxonomy: ModelTaxonomyMetadata
        The model taxonomy metadata.
    model_artifact: ModelArtifact
        This is built by calling prepare.
    model_deployment: ModelDeployment
        A ModelDeployment instance.
    model_file_name: str
        Name of the serialized model.
    model_id: str
        The model ID.
    properties: ModelProperties
        ModelProperties object required to save and deploy model.
        For more details, check https://accelerated-data-science.readthedocs.io/en/latest/ads.model.html#module-ads.model.model_properties.
    runtime_info: RuntimeInfo
        A RuntimeInfo instance.
    schema_input: Schema
        Schema describes the structure of the input data.
    schema_output: Schema
        Schema describes the structure of the output data.
    serialize: bool
        Whether to serialize the model to pkl file by default. If False, you need to serialize the model manually,
        save it under artifact_dir and update the score.py manually.
    version: str
        The framework version of the model.

    Methods
    -------
    delete_deployment(...)
        Deletes the current model deployment.
    deploy(..., **kwargs)
        Deploys a model.
    from_model_artifact(uri, model_file_name, artifact_dir, ..., **kwargs)
        Loads model from the specified folder, or zip/tar archive.
    from_model_catalog(model_id, model_file_name, artifact_dir, ..., **kwargs)
        Loads model from model catalog.
    introspect(...)
        Runs model introspection.
    predict(data, ...)
        Returns prediction of input data run against the model deployment endpoint.
    prepare(..., **kwargs)
        Prepare and save the score.py, serialized model and runtime.yaml file.
    reload(...)
        Reloads the model artifact files: `score.py` and the `runtime.yaml`.
    save(..., **kwargs)
        Saves model artifacts to the model catalog.
    summary_status(...)
        Gets a summary table of the current status.
    verify(data, ...)
        Tests if deployment works in local environment.

    Examples
    --------
    >>> torch_model = PyTorchModel(estimator=torch_estimator,
    ... artifact_dir=tmp_model_dir)
    >>> inference_conda_env = "generalml_p37_cpu_v1"

    >>> torch_model.prepare(inference_conda_env=inference_conda_env, force_overwrite=True)
    >>> torch_model.reload()
    >>> torch_model.verify(...)
    >>> torch_model.save()
    >>> model_deployment = torch_model.deploy(wait_for_completion=False)
    >>> torch_model.predict(...)
    """

    _PREFIX = "pytorch"
    model_save_serializer_type = PyTorchModelSerializerType

    @runtime_dependency(module="torch", install_from=OptionalDependency.PYTORCH)
    def __init__(
        self,
        estimator: callable,
        artifact_dir: str,
        properties: Optional[ModelProperties] = None,
        auth: Dict = None,
        model_save_serializer: Optional[SERDE] = model_save_serializer_type.TORCH,
        model_input_serializer: Optional[SERDE] = None,
        **kwargs,
    ):
        """
        Initiates a PyTorchModel instance.

        Parameters
        ----------
        estimator: callable
            Any model object generated by pytorch framework
        artifact_dir: str
            artifact directory to store the files needed for deployment.
        properties: (ModelProperties, optional). Defaults to None.
            ModelProperties object required to save and deploy model.
        auth :(Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        model_save_serializer: (SERDE or str, optional). Defaults to None.
            Instance of ads.model.SERDE. Used for serialize/deserialize model.
        model_input_serializer: (SERDE, optional). Defaults to None.
            Instance of ads.model.SERDE. Used for serialize/deserialize data.

        Returns
        -------
        PyTorchModel
            PyTorchModel instance.
        """
        super().__init__(
            estimator=estimator,
            artifact_dir=artifact_dir,
            properties=properties,
            auth=auth,
            model_save_serializer=model_save_serializer,
            model_input_serializer=model_input_serializer,
            **kwargs,
        )
        self._extractor = PytorchExtractor(estimator)
        self.framework = self._extractor.framework
        self.algorithm = self._extractor.algorithm
        self.version = self._extractor.version
        self.hyperparameter = self._extractor.hyperparameter
        self.version = torch.__version__

    def serialize_model(
        self,
        as_onnx: bool = False,
        force_overwrite: bool = False,
        X_sample: Optional[
            Union[
                Dict,
                str,
                List,
                Tuple,
                np.ndarray,
                pd.core.series.Series,
                pd.core.frame.DataFrame,
            ]
        ] = None,
        use_torch_script: bool = None,
        **kwargs,
    ) -> None:
        """
        Serialize and save Pytorch model using ONNX or model specific method.

        Parameters
        ----------
        as_onnx: (bool, optional). Defaults to False.
            If set as True, convert into ONNX model.
        force_overwrite: (bool, optional). Defaults to False.
            If set as True, overwrite serialized model if exists.
        X_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]. Defaults to None.
            A sample of input data that will be used to generate input schema and detect onnx_args.
        use_torch_script:  (bool, optional). Defaults to None (If the default value has not been changed, it will be set as `False`).
            If set as `True`, the model will be serialized as a TorchScript program. Check https://pytorch.org/tutorials/beginner/saving_loading_models.html#export-load-model-in-torchscript-format for more details.
            If set as `False`, it will only save the trained model’s learned parameters, and the score.py
            need to be modified to construct the model class instance first. Check https://pytorch.org/tutorials/beginner/saving_loading_models.html#save-load-state-dict-recommended for more details.
        **kwargs: optional params used to serialize pytorch model to onnx,
        including the following:
            onnx_args: (tuple or torch.Tensor), default to None
            Contains model inputs such that model(onnx_args) is a valid
            invocation of the model. Can be structured either as: 1) ONLY A
            TUPLE OF ARGUMENTS; 2) A TENSOR; 3) A TUPLE OF ARGUMENTS ENDING
            WITH A DICTIONARY OF NAMED ARGUMENTS
            input_names: (List[str], optional). Names to assign to the input
            nodes of the graph, in order.
            output_names: (List[str], optional). Names to assign to the output nodes of the graph, in order.
            dynamic_axes: (dict, optional), default to None. Specify axes of tensors as dynamic (i.e. known only at run-time).

        Returns
        -------
        None
            Nothing.
        """
        if use_torch_script is None:
            logger.warning(
                "In future the models will be saved in TorchScript format by default. Currently saving it using torch.save method."
                "Set `use_torch_script` as `True` to serialize the model as a TorchScript program by `torch.jit.save()` "
                "and loaded using `torch.jit.load()` in score.py. "
                "You don't need to modify `load_model()` in score.py to load the model."
                "Check https://pytorch.org/tutorials/beginner/saving_loading_models.html#export-load-model-in-torchscript-format for more details."
                "Set `use_torch_script` as `False` to save only the model parameters."
                "The model class instance must be constructed before "
                "loading parameters in the perdict function of score.py."
                "Check https://pytorch.org/tutorials/beginner/saving_loading_models.html#save-load-state-dict-recommended for more details."
            )
            use_torch_script = False

        if as_onnx and use_torch_script:
            raise ValueError("You can only save Pytorch model into one format.")

        if as_onnx:
            logger.warning(DEPRECATE_AS_ONNX_WARNING)
            self.set_model_save_serializer(self.model_save_serializer_type.ONNX)

        if use_torch_script:
            logger.warning(DEPRECATE_USE_TORCH_SCRIPT_WARNING)
            self.set_model_save_serializer(self.model_save_serializer_type.TORCHSCRIPT)

        super().serialize_model(
            as_onnx=as_onnx,
            force_overwrite=force_overwrite,
            X_sample=X_sample,
            **kwargs,
        )

    def _to_tensor(self, data):
        try:
            import torchvision.transforms as transforms

            convert_tensor = transforms.ToTensor()
            data = convert_tensor(data)
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                f"The `torchvision` module was not found. Please run "
                f"`pip install {OptionalDependency.PYTORCH}`."
            )
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import json
from typing import Dict, List

import oci.ai_language
import pandas as pd
from ads.common import auth as authutil
from ads.common import oci_client


class OCILanguage(object):  # pragma: no cover
    """Defines the OCILanguage plugin for ADSString built on top of the OCI Language Services.

    Example
    -------
    >>> ADSString.plugin_register(OCILanguage)
    >>> s = ADSString("This movie is awesome.")
    >>> s.absa
    >>> s.text_classification
    >>> s.ner
    >>> s.language_dominant
    """

    def __init__(self, auth=None):
        auth = auth if auth else authutil.default_signer()
        self.ai_client = oci_client.OCIClientFactory(**auth).ai_language

    @property
    def ner(self) -> List[Dict]:
        """Detects named entites in the text."""
        output = self.ai_client.batch_detect_language_entities(
            oci.ai_language.models.BatchDetectLanguageEntitiesDetails(
                documents=[
                    oci.ai_language.models.TextDocument(key="1", text=self.string)
                ]
            )
        )
        return json.loads(str(output.data.documents[0]))["entities"]

    @property
    def language_dominant(self) -> List[Dict]:
        """Determines the language of the text along with ISO 639-1 language code and a probability score."""
        output = self.ai_client.batch_detect_dominant_language(
            oci.ai_language.models.BatchDetectDominantLanguageDetails(
                documents=[
                    oci.ai_language.models.DominantLanguageDocument(
                        key="1", text=self.string
                    )
                ]
            )
        )
        return json.loads(str(output.data.documents[0]))

    @property
    def key_phrase(self) -> List[Dict]:
        """Extracts the most relevant words from the ADSString object and assigns them a score."""
        output = self.ai_client.batch_detect_language_key_phrases(
            oci.ai_language.models.BatchDetectLanguageKeyPhrasesDetails(
                documents=[
                    oci.ai_language.models.TextDocument(key="1", text=self.string)
                ]
            )
        )
        return json.loads(str(output.data.documents[0]))["key_phrases"]

    @property
    def absa(self) -> List[Dict]:
        """Runs aspect-based sentiment analysis on the text to gauge teh mood or the tone of the text."""
        output = self.ai_client.batch_detect_language_sentiments(
            oci.ai_language.models.BatchDetectLanguageSentimentsDetails(
                documents=[
                    oci.ai_language.models.TextDocument(key="1", text=self.string)
                ]
            )
        )
        return json.loads(str(output.data.documents[0]))["aspects"]

    @property
    def text_classification(self) -> List[Dict]:
        """Analyses the text and identifies categories for the content with a confidence score."""
        output = self.ai_client.batch_detect_language_text_classification(
            oci.ai_language.models.BatchDetectLanguageTextClassificationDetails(
                documents=[
                    oci.ai_language.models.TextDocument(key="1", text=self.string)
                ]
            )
        )
        return json.loads(str(output.data.documents[0]))["text_classification"]



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents a String feature type.

Classes:
    String
        The feature type that represents string values.
"""
import matplotlib.pyplot as plt
import pandas as pd
from ads.feature_engineering.feature_type.base import FeatureType
from ads.feature_engineering.utils import (
    _count_unique_missing,
    random_color_func,
    SchemeNeutral,
)
from ads.feature_engineering import schema
from ads.common import utils, logger
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


def default_handler(data: pd.Series, *args, **kwargs) -> pd.Series:
    """Processes given data and indicates if the data matches requirements.

    Parameters
    ----------
    data: pd.Series
        The data to process.

    Returns
    -------
    pd.Series: The logical list indicating if the data matches requirements.
    """
    return data.apply(lambda x: isinstance(x, str))


class String(FeatureType):
    """
    Type representing string values.

    Attributes
    ----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.

    Methods
    --------
    feature_stat(x: pd.Series) -> pd.DataFrame
        Generates feature statistics.
    feature_plot(x: pd.Series) -> plt.Axes
        Shows distributions of datasets using wordcloud.


    Example
    -------
    >>> from ads.feature_engineering.feature_type.string import String
    >>> import pandas as pd
    >>> s = pd.Series(["Hello", "world", None], name='string')
    >>> String.validator.is_string(s)
    0     True
    1     True
    2    False
    Name: string, dtype: bool
    """

    description = "Type representing string values."

    @staticmethod
    def feature_stat(x: pd.Series) -> pd.DataFrame:
        """Generates feature statistics.

        Feature statistics include (total)count, unique(count)
        and missing(count) if there is any.

        Examples
        --------
        >>> string = pd.Series(['S', 'C', 'S', 'S', 'S', 'Q', 'S', 'S', 'S', 'C', 'S', 'S', 'S',
                'S', 'S', 'S', 'Q', 'S', 'S', '', np.NaN, None], name='string')
        >>> string.ads.feature_type = ['string']
        >>> string.ads.feature_stat()
            Metric  Value
        0	count	22
        1	unique	3
        2	missing	3

        Returns
        -------
        Pandas Dataframe
            Summary statistics of the Series or Dataframe provided.
        """
        return _count_unique_missing(x)

    @staticmethod
    @runtime_dependency(module="wordcloud", install_from=OptionalDependency.TEXT)
    def feature_plot(x: pd.Series) -> plt.Axes:
        """
        Shows distributions of datasets using wordcloud.

        Examples
        --------
        >>> string = pd.Series(['S', 'C', 'S', 'S', 'S', 'Q', 'S', 'S', 'S', 'C', 'S', 'S', 'S',
                'S', 'S', 'S', 'Q', 'S', 'S', '', np.NaN, None], name='string')
        >>> string.ads.feature_type = ['string']
        >>> string.ads.feature_plot()

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
            Plot object for the series based on the String feature type.
        """
        col_name = x.name if x.name else "text"
        df = x.to_frame(col_name)
        df["validation"] = default_handler(x)
        df = df[df["validation"] == True]
        words = " ".join(df[col_name].dropna().to_list())
        if not words:
            return

        from wordcloud import WordCloud

        wc = WordCloud(
            background_color=SchemeNeutral.BACKGROUND_LIGHT,
            color_func=random_color_func,
        ).generate(words)
        _, ax = plt.subplots(facecolor=SchemeNeutral.BACKGROUND_LIGHT)
        ax.imshow(wc)
        plt.axis("off")
        return ax

    @classmethod
    def feature_domain(cls, x: pd.Series) -> schema.Domain:
        """
        Generate the domain of the data of this feature type.

        Examples
        --------
        >>> string = pd.Series(['S', 'C', 'S', 'S', 'S', 'Q', 'S', 'S', 'S', 'C', 'S', 'S', 'S',
                'S', 'S', 'S', 'Q', 'S', 'S', '', np.NaN, None], name='string')
        >>> string.ads.feature_type = ['string']
        >>> string.ads.feature_domain()
        constraints: []
        stats:
            count: 22
            missing: 3
            unique: 3
        values: String

        Returns
        -------
        ads.feature_engineering.schema.Domain
            Domain based on the String feature type.
        """

        return schema.Domain(cls.__name__, cls.feature_stat(x).to_dict()[x.name], [])


String.validator.register("is_string", default_handler)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from ads.model.extractor.model_info_extractor import (
    ModelInfoExtractor,
    normalize_hyperparameter,
)
from ads.model.model_metadata import Framework
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


class HuggingFaceExtractor(ModelInfoExtractor):
    """Class that extract model metadata from HuggingFacePipeline models.

    Attributes
    ----------
    model: object
        The model to extract metadata from.
    estimator: object
        The estimator to extract metadata from.

    Methods
    -------
    framework(self) -> str
        Returns the framework of the model.
    algorithm(self) -> object
        Returns the algorithm of the model.
    version(self) -> str
        Returns the version of framework of the model.
    hyperparameter(self) -> dict
        Returns the hyperparameter of the model.
    """

    def __init__(self, model):
        self.model = model

    @property
    def framework(self):
        """Extracts the framework of the model.

        Returns
        ----------
        str:
           The framework of the model.
        """
        return Framework.TRANSFORMERS

    @property
    def algorithm(self):
        """Extracts the algorithm of the model.

        Returns
        ----------
        object:
           The algorithm of the model.
        """
        return self.model.__class__.__name__

    @property
    @runtime_dependency(
        module="transformers", install_from=OptionalDependency.HUGGINGFACE
    )
    def version(self):
        """Extracts the framework version of the model.

        Returns
        ----------
        str:
           The framework version of the model.
        """
        return transformers.__version__

    @property
    def hyperparameter(self):
        """Extracts the hyperparameters of the model.

        Returns
        ----------
        dict:
           The hyperparameters of the model.
        """
        return normalize_hyperparameter(self.model.model.config.to_dict())



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from enum import Enum

DEFAULT_OCI_CONFIG_FILE = "~/.oci/config"
DEFAULT_PROFILE = "DEFAULT"
DEFAULT_CONDA_PACK_FOLDER = "~/conda"
CONDA_PACK_OS_PREFIX_FORMAT = "oci://<bucket>@<namespace>/<prefix>"
DEFAULT_ADS_CONFIG_FOLDER = "~/.ads_ops"
OPS_IMAGE_BASE = "ads-operators-base"
ML_JOB_IMAGE = "ml-job"
ML_JOB_GPU_IMAGE = "ml-job-gpu"
OPS_IMAGE_GPU_BASE = "ads-operators-gpu-base"
DEFAULT_MANIFEST_VERSION = "1.0"
ADS_CONFIG_FILE_NAME = "config.ini"
ADS_JOBS_CONFIG_FILE_NAME = "ml_job_config.ini"
ADS_DATAFLOW_CONFIG_FILE_NAME = "dataflow_config.ini"
ADS_ML_PIPELINE_CONFIG_FILE_NAME = "ml_pipeline.ini"
ADS_LOCAL_BACKEND_CONFIG_FILE_NAME = "local_backend.ini"
DEFAULT_IMAGE_HOME_DIR = "/home/datascience"
DEFAULT_IMAGE_SCRIPT_DIR = "/etc/datascience"
DEFAULT_IMAGE_CONDA_DIR = "/opt/conda/envs"
DEFAULT_NOTEBOOK_SESSION_SPARK_CONF_DIR = "/home/datascience/spark_conf_dir"
DEFAULT_NOTEBOOK_SESSION_CONDA_DIR = "/home/datascience/conda"


class BACKEND_NAME(Enum):
    JOB = "job"
    DATAFLOW = "dataflow"
    PIPELINE = "pipeline"
    MODEL_DEPLOYMENT = "deployment"
    LOCAL = "local"



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import itertools
from typing import Any, Callable, Dict, Generator, List, Union

import ads
import ads.text_dataset.extractor as te
import fsspec
import pandas as pd
from ads.text_dataset import backends
from ads.text_dataset.options import OptionFactory, Options
from ads.text_dataset.udfs import UDF
from ads.text_dataset.utils import NotSupportedError


class DataLoader:
    """
    DataLoader binds engine, FileProcessor and File handler(in this case it is fsspec)
    together to produce a dataframe of parsed text from files.

    This class is expected to be used mainly from TextDatasetFactory class.

    Attributes
    ----------
    processor: `ads.text_dataset.extractor.FileProcessor`
        processor that is used for loading data.

    Examples
    --------
    >>> import oci
    >>> from ads.text_dataset.dataset import TextDatasetFactory as textfactory
    >>> from ads.text_dataset.options import Options
    >>> df = textfactory.format('pdf').engine('pandas').read_line(
    ...     'oci://<bucket-name>@<namespace>/<path>/*.pdf',
    ...     storage_options={"config": oci.config.from_file(os.path.join("~/.oci", "config"))},
    ... )
    >>> data_gen = textfactory.format('pdf').option(Options.FILE_NAME).backend('pdfplumber').read_text(
    ...     'oci://<bucket-name>@<namespace>/<path>/*.pdf',
    ...     storage_options={"config": oci.config.from_file(os.path.join("~/.oci", "config"))},
    ... )
    >>> textfactory.format('docx').convert_to_text(
    ...     'oci://<bucket-name>@<namespace>/<path>/*.docx',
    ...     './extracted',
    ...     storage_options={"config": oci.config.from_file(os.path.join("~/.oci", "config"))},
    ... )
    >>> textfactory.format('docx').convert_to_text(
    ...     'oci://<bucket-name>@<namespace>/<path>/*.docx',
    ...     'oci://<bucket-name>@<namespace>/<out_path>',
    ...     storage_options={"config": oci.config.from_file(os.path.join("~/.oci", "config"))},
    ... )
    >>> meta_gen = textfactory.format('docx').metadata_schema(
    ...     'oci://<bucket-name>@<namespace>/papers/*.pdf',
    ...     storage_options={"config": oci.config.from_file(os.path.join("~/.oci", "config"))},
    ... )
    >>> df = textfactory.format('pdf').engine('pandas').option(Options.FILE_METADATA, {'extract': ['Author']}).read_text(
    ...     'oci://<bucket-name>@<namespace>/<path>/*.pdf',
    ...     storage_options={"config": oci.config.from_file(os.path.join("~/.oci", "config"))},
    ...     total_files=10,
    ... )
    >>> df = textfactory.format('txt').engine('cudf').read_line(
    ...     'oci://<bucket-name>@<namespace>/<path>/*.log',
    ...      udf=r'^\[(\S+)\s(\S+)\s(\d+)\s(\d+\:\d+\:\d+)\s(\d+)]\s(\S+)\s(\S+)\s(\S+)\s(\S+)',
    ...      df_args={"columns":["day", "month", "date", "time", "year", "type", "method", "status", "file"]},
    ...      n_lines_per_file=10,
    ... )
    """

    def __init__(self, engine: str = None) -> None:
        """Initialize a DataLoader object.

        Parameters
        ----------
        engine : str, optional
            dataframe engine, by default None.

        Returns
        -------
        None
        """
        self.engine(engine)
        self.filemanager = fsspec
        self.processor = te.FileProcessorFactory.get_processor("txt")
        self.options = []
        self._data = None

    def with_processor(self, processor_type: str) -> None:
        """Set file processor.

        Parameters
        ----------
        processor_type : str
            type of processor, which corresponds to format of the file.

        Returns
        -------
        None
        """
        self.processor = te.FileProcessorFactory.get_processor(processor_type)()
        return self

    def engine(self, eng: str) -> None:
        """Set engine for dataloader. Can be pandas or cudf.

        Parameters
        ----------
        eng : str
            name of engine

        Returns
        -------
        None

        Raises
        ------
        NotSupportedError
            raises error if engine passed in is not supported.
        """
        if eng is None:
            self._engine = None
            self._format_output = lambda *args, **kwargs: args[0]
            return self
        if eng not in ["pandas", "cudf"]:
            raise NotSupportedError("Only pandas and cudf currently.")
        else:
            if eng == "pandas":
                import pandas

                self._engine = pandas
                self._format_output = pandas.DataFrame
            else:
                import cudf

                self._engine = cudf
                self._format_output = lambda data, **kwargs: cudf.DataFrame(
                    [row for row in data], **kwargs
                )  # cuDF cannot be initialized with a generator
        return self

    def backend(self, backend: Union[str, backends.Base]) -> None:
        """Set backend used for extracting text from files.

        Parameters
        ----------
        backend : (str | `ads.text_dataset.backends.Base`)
            backend for extracting text from raw files.

        Returns
        -------
        None
        """
        self.processor.backend(backend)
        return self

    def option(self, opt: Options, spec: Any = None) -> None:
        """Set extraction options.

        Parameters
        ----------
        opt : `ads.text_dataset.options.Options`
            an option defined in `ads.text_dataset.options.Options`
        spec : Any, optional
            specifications that will be passed to option handler, by default None

        Returns
        -------
        None
        """
        self.options.append((OptionFactory.option_handler(opt), spec))
        return self

    def __load_data__(
        self,
        reader: Callable,
        path: str,
        udf: Union[str, Callable] = None,
        storage_options: Dict = None,
        encoding: str = "utf-8",
        n_rows_per_file: int = None,
        total_rows: int = None,
    ) -> Generator[Union[str, List[str]], None, None]:
        storage_options = storage_options if storage_options is not None else {}
        fhs = self.filemanager.open_files(
            path, mode="rb", encoding=encoding, **storage_options
        )
        if udf is not None:
            if isinstance(udf, str):
                fn = UDF.from_regex(udf)
            else:
                fn = udf
        else:
            fn = lambda x: x

        total_line_count = [0]

        # function to apply to each element
        def func(fh, reader):
            out = [option(self).handle(fh, spec) for option, spec in self.options]
            line_count = 0
            for text in reader(fh):
                if total_rows is None or total_line_count[0] < total_rows:
                    if n_rows_per_file is None or line_count < n_rows_per_file:
                        content = fn(text)
                        if content is not None:
                            yield out + list(content) if (
                                isinstance(content, list) or isinstance(content, tuple)
                            ) else out + [content]
                            line_count += 1
                            total_line_count[0] += 1

        return itertools.chain.from_iterable((func(fh, reader) for fh in fhs))

    def read_line(
        self,
        path: str,
        udf: Union[str, Callable] = None,
        n_lines_per_file: int = None,
        total_lines: int = None,
        df_args: Dict = None,
        storage_options: Dict = None,
        encoding: str = "utf-8",
    ) -> Union[Generator[Union[str, List[str]], None, None], "DataFrame"]:
        """Read each file into lines. If path matches multiple files, will combine lines from all files.

        Parameters
        ----------
        path : str
            path to data files. can have glob pattern.
        udf : (callable | str), optional
            user defined function for processing each line, can be a callable or regex, by default None
        n_lines_per_file : int, optional
            max number of lines read from each file, by default None
        total_lines : int, optional
            max number of lines read from all files, by default None
        df_args : dict, optional
            arguments passed to dataframe engine (e.g. pandas), by default None
        storage_options : dict, optional
            storage options for cloud storage, by default None
        encoding : str, optional
            encoding of files, by default 'utf-8'

        Returns
        -------
        (Generator | DataFrame)
            returns either a data generator or a dataframe.
        """
        df_args = df_args if df_args is not None else {}
        self._data = self.__load_data__(
            self.processor.read_line,
            path,
            udf,
            storage_options,
            encoding,
            n_lines_per_file,
            total_lines,
        )
        return self._format_output(self._data, **df_args)

    def read_text(
        self,
        path: str,
        udf: Union[str, Callable] = None,
        total_files: int = None,
        storage_options: Dict = None,
        df_args: Dict = None,
        encoding: str = "utf-8",
    ) -> Union[Generator[Union[str, List[str]], None, None], "DataFrame"]:
        """Read each file into a text string. If path matches multiple files, each file corresponds to one record.

        Parameters
        ----------
        path : str
            path to data files. can have glob pattern.
        udf : (callable | str), optional
            user defined function for processing each line, can be a callable or regex, by default None
        total_files : int, optional
            max number of files to read, by default None
        df_args : dict, optional
            arguments passed to dataframe engine (e.g. pandas), by default None
        storage_options : dict, optional
            storage options for cloud storage, by default None
        encoding : str, optional
            encoding of files, by default 'utf-8'

        Returns
        -------
        (Generator | DataFrame)
            returns either a data generator or a dataframe.
        """
        df_args = df_args if df_args is not None else {}
        self._data = self.__load_data__(
            self.processor.read_text,
            path,
            udf,
            storage_options,
            encoding,
            1,
            total_files,
        )
        return self._format_output(self._data, **df_args)

    def convert_to_text(
        self,
        src_path: str,
        dst_path: str,
        encoding: str = "utf-8",
        storage_options: Dict = None,
    ) -> None:
        """Convert files to plain text files.

        Parameters
        ----------
        src_path : str
            path to source data file(s). can use glob pattern
        dst_path: str
            local folder or cloud storage (e.g., OCI object storage) prefix to save converted text files
        encoding: str, optional
            encoding for files, by default utf-8
        storage_options : Dict, optional
            storage options for cloud storage, by default None

        Returns
        -------
        None
        """
        storage_options = storage_options if storage_options is not None else {}
        fhs = self.filemanager.open_files(
            src_path, mode="rb", encoding=encoding, **storage_options
        )
        for fh in fhs:
            self.processor.convert_to_text(
                fh,
                dst_path,
                storage_options=storage_options,
            )

    def metadata_all(
        self, path: str, storage_options: Dict = None, encoding: str = "utf-8"
    ) -> Generator[Dict[str, Any], None, None]:
        """Get metadata of all files that matches the given path. Return a generator.

        Parameters
        ----------
        path : str
            path to data files. can use glob pattern.
        storage_options : Dict, optional
            storage options for cloud storage, by default None
        encoding : str, optional
            encoding of files, by default 'utf-8'

        Returns
        -------
        Generator
            generator of extracted metedata from files.
        """
        storage_options = storage_options if storage_options is not None else {}
        fhs = self.filemanager.open_files(
            path, mode="rb", encoding=encoding, **storage_options
        )
        return (self.processor.get_metadata(fh) for fh in fhs)

    def metadata_schema(
        self,
        path: str,
        n_files: int = 1,
        storage_options: Dict = None,
        encoding: str = "utf-8",
    ) -> List[str]:
        """
        Get available fields in metadata by looking at the first `n_files` that
        matches the given path.

        Parameters
        ----------
        path: str
            path to data files. can have glob pattern
        n_files: int, optional
            number of files to look up, default to be 1
        storage_options: dict, optional
            storage options for cloud storage, by default None
        encoding: str, optional
            encoding of files, by default utf-8

        Returns
        -------
        List[str]
            list of available fields in metadata
        """

        metadata = self.metadata_all(
            path, storage_options=storage_options, encoding=encoding
        )
        fields = set()
        for _ in range(n_files):
            try:
                fields.update(list(next(metadata).keys()))
            except StopIteration:
                break
        return list(fields)

    # ----- not currently used, but in case we want to consider chaining in the future -----
    def _transform(self, udf, udf_type="fn"):  # pragma: no cover
        if udf_type == "fn":
            func = UDF.from_lambda(udf)
        elif udf_type == "regex":
            func = UDF.from_regex(udf)
        else:
            raise NotImplementedError("Other types of UDF not yet supported.")

        # convert df into iterator
        if isinstance(self._data, pd.DataFrame) or isinstance(self._data, pd.Series):
            self._data = (
                row.values if len(row.values) > 1 else row.values[0]
                for i, row in self._data.iterrows()
            )

        self._data = (func(row) for row in self._data)
        self._data = (row for row in self._data if row is not None)
        return self


class TextDatasetFactory:
    """A class that generates a dataloader given a file format."""

    @staticmethod
    def format(format_name: str) -> DataLoader:
        """
        Instantiates DataLoader class and seeds it with the right kind of FileProcessor.
        Eg. PDFProcessor for pdf. The FileProcessorFactory returns the processor based
        on the format Type.

        Parameters
        ----------
        format_name : str
            name of format

        Returns
        -------
        `ads.text_dataset.dataset.DataLoader`
            a `DataLoader` object.
        """
        return DataLoader().with_processor(format_name)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import yaml
import os
from ads.opctl.distributed.common.abstract_cluster_provider import ClusterProvider
from ads.opctl.diagnostics.check_distributed_job_requirements import (
    PortsChecker,
    Default,
)
from ads.opctl.diagnostics.requirement_exception import RequirementException
from jinja2 import Environment, PackageLoader
import fsspec
import importlib
import sys

cp = ClusterProvider(mode="MAIN")
auth = cp.get_oci_auth()


class RequirementChecker:

    providers = {}

    CHECK = "\u2705"
    CROSS = "\u274E"
    TADA = "\u2728"
    DIAGNOSTIC_REPORT_FILE = "diagnostic_report.html"

    @classmethod
    def register_provider(cls, key: str, value):
        RequirementChecker.providers[key] = value

    def load_requirement(self, framework):
        req = {}
        with open(f"{framework}_req.yaml") as rf:
            req = yaml.load(rf, Loader=yaml.SafeLoader)
        return req

    def setup_syspath(self):
        code_dir = os.environ.get("OCI__CODE_DIR", "/code/")
        sys.path.append(code_dir)
        for dirname in os.listdir(code_dir):
            file = os.path.join(code_dir, dirname)
            if os.path.isdir(file):
                sys.path.append(file)

    def load_external_modules(self, modules):
        for module in modules:
            importlib.import_module(module)
        print(f"Loaded modules: {modules}")

    def verify_requirements(self, kind):

        self.kind = kind
        self.setup_syspath()
        configuration = {}
        requirements = {}
        cluster_type = os.environ["OCI__CLUSTER_TYPE"].lower()
        DIAGNOSTIC_CONFIG = os.environ.get(
            "OCI__DIAGNOSTIC_DEFAULT_CONFIG"
        ) or os.path.join(
            os.path.dirname(os.path.dirname(__file__)),
            "config",
            "diagnostics",
            self.kind,
            "default_requirements_config.yaml",
        )
        print(f"Loading requirements for diagnostics from {DIAGNOSTIC_CONFIG}")
        with open(DIAGNOSTIC_CONFIG) as cf:
            configuration = yaml.load(cf, Loader=yaml.SafeLoader)[cluster_type]
            requirements = configuration.get("checkers")
        if configuration.get("checker_provider"):
            self.load_external_modules(configuration["checker_provider"])

        failures = []
        passed = []
        for requirement in requirements:
            print(f"Checking requirement: {requirement}")
            provider = ".".join(requirement["checker"].split(".")[:-1])
            method = requirement["checker"].split(".")[-1]
            kwargs = requirement.get("args", {})
            try:
                getattr(RequirementChecker.providers[provider](auth=auth), method)(
                    **kwargs
                )
                passed.append(requirement["description"])
            except RequirementException as re:
                failures.append((requirement["description"], str(re)))

        for item in passed:
            print(f"{item}: {RequirementChecker.CHECK}", flush=True)
        for item in failures:
            print(
                f"{item[0]}: {RequirementChecker.CROSS} . Resolution: {item[1]}",
                flush=True,
            )
        if len(failures) == 0:
            print(
                f"{RequirementChecker.TADA} You have satisfied all the requirements for cluster: {cluster_type} \u2728",
                flush=True,
            )
        self.generate_report(cluster_type, passed, failures)

    def generate_report(self, cluster_type, passed, failures):
        _env = Environment(loader=PackageLoader("ads", "opctl/templates"))
        # import jinja2

        # loader = jinja2.FileSystemLoader(".")
        # _env = jinja2.Environment(loader=loader)
        report_template = _env.get_template(f"diagnostic_report_template.jinja2")
        report = [(req, RequirementChecker.CHECK) for req in passed] + [
            (req[0], RequirementChecker.CROSS) for req in failures
        ]
        success_message = f"{RequirementChecker.TADA} You have satisfied all the requirements for cluster: {cluster_type} {RequirementChecker.TADA}"
        report_path = os.path.join(
            os.path.join(os.environ.get("OCI__WORK_DIR", ""), cp.my_work_dir),
            RequirementChecker.DIAGNOSTIC_REPORT_FILE,
        )
        with fsspec.open(
            # "diagnostic_report.html",
            report_path,
            "w",
            **auth,
        ) as of:
            of.write(
                report_template.render(
                    cluster_type=cluster_type,
                    report=report,
                    action_item=failures,
                    success_message=success_message,
                )
            )
            print(f"Finished writing to {report_path}")


if __name__ == "__main__":
    # fetch_security_list(subnet_id)
    RequirementChecker.register_provider("cluster-ports-check", PortsChecker)
    RequirementChecker.register_provider("cluster-generic", Default)
    RequirementChecker().verify_requirements()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import json
import ads
from ads.secrets import SecretKeeper, Secret
from dataclasses import dataclass


@dataclass
class AuthToken(Secret):
    """
    AuthToken dataclass holds `auth_token` attribute
    """

    auth_token: str


class AuthTokenSecretKeeper(SecretKeeper):
    """
    `AuthTokenSecretKeeper` uses `ads.secrets.auth_token.AuthToken` class to manage Auth Token credentials.
    The credentials are stored in Vault as a dictionary with the following format - `{"auth_token":"user provided value"}`

    Examples
    --------

    >>> from ads.secrets.auth_token import AuthTokenSecretKeeper
    >>> import ads
    >>> ads.set_auth("resource_principal") #If using resource principal for authentication
    >>> # Save Auth Tokens or Acess Keys to the vault
    >>>
    >>>
    >>> authtoken2 = AuthTokenSecretKeeper(vault_id=vault_id,
    ...            key_id=key_id,
    ...            auth_token="<your auth token>").save("my_xyz_auth_token2",
    ...                                                                 "This is my auth token for git repo xyz",
    ...                                                                 freeform_tags={"gitrepo":"xyz"})
    >>> authtoken2.export_vault_details("my_git_token_vault_info.yaml", format="yaml")
    >>> # Loading credentials
    >>> with AuthTokenSecretKeeper.load_secret(source="ocid1.vaultsecret.oc1..<unique_ID>",
    ...                                export_prefix="mygitrepo",
    ...                                export_env=True
    ...                               ) as authtoken:
    ...     import os
    ...     print("Credentials inside environment variable:", os.environ.get('mygitrepo.auth_token'))
    ...     print("Credentials inside `authtoken` object: ", authtoken)
    Credentials inside environment variable: <your auth token>
    Credentials inside `authtoken` object:  {'auth_token': '<your auth token>'}
    >>> print("Credentials inside `authtoken` object: ", authtoken)
    Credentials inside `authtoken` object:  {'auth_token': None}
    >>> print("Credentials inside environment variable:", os.environ.get('mygitrepo.auth_token'))
    Credentials inside environment variable: None

    """

    def __init__(self, auth_token=None, **kwargs):
        """
        Parameters
        ----------
        auth_token: (str, optional). Default None
            auth token string that needs to be stored in the vault
        kwargs:
            vault_id: str. OCID of the vault where the secret is stored. Required for saving secret.
            key_id: str. OCID of the key used for encrypting the secret. Required for saving secret.
            compartment_id: str. OCID of the compartment where the vault is located. Required for saving secret.
            auth: dict. Dictionay returned from ads.common.auth.api_keys() or ads.common.auth.resource_principal(). By default, will follow what is set in `ads.set_auth`.  Use this attribute to override the default.

        """
        self.data = AuthToken(auth_token)
        super().__init__(**kwargs)

    def decode(self) -> "ads.secrets.auth_token.AuthTokenSecretKeeper":
        """
        Converts the content in `self.encoded` to `AuthToken` and stores in `self.data`

        Returns
        -------
        AuthTokenSecretKeeper:
            Returns the self object after decoding `self.encoded` and updates `self.data`
        """
        content = json.loads(self._decode())
        self.data = AuthToken(**content)
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


class InvalidFeatureType(TypeError):
    def __init__(self, tname: str):
        super().__init__(f"Type {tname} is not a sublcass of FeatureType.")


class TypeAlreadyRegistered(TypeError):
    def __init__(self, tname: str):
        super().__init__(f"Type {tname} already registered.")


class TypeAlreadyAdded(TypeError):
    def __init__(self, tname: str):
        super().__init__(f"Type {tname} is already added.")


class TypeNotFound(TypeError):
    def __init__(self, tname: str):
        super().__init__(f"Type {tname} is not found.")


class NameAlreadyRegistered(NameError):
    def __init__(self, name: str):
        super().__init__(f"Type with name {name} already registered.")


class WarningAlreadyExists(ValueError):
    def __init__(self, name: str):
        super().__init__(f"Warning {name} already exists.")


class WarningNotFound(ValueError):
    def __init__(self, name: str):
        super().__init__(f"Warning {name} is not found.")



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
import shutil
import tempfile
import uuid
from abc import ABC, abstractmethod
from typing import Dict, Optional
from zipfile import ZipFile

from ads.common import utils
from ads.common.utils import extract_region
from ads.model.service.oci_datascience_model import OCIDataScienceModel


class ArtifactDownloader(ABC):
    """The abstract class to download model artifacts."""

    PROGRESS_STEPS_COUNT = 1

    def __init__(
        self,
        dsc_model: OCIDataScienceModel,
        target_dir: str,
        force_overwrite: Optional[bool] = False,
    ):
        """Initializes `ArtifactDownloader` instance.

        Parameters
        ----------
        dsc_model: OCIDataScienceModel
            The data scince model instance.
        target_dir: str
            The target location of model after download.
        force_overwrite: bool
            Overwrite target_dir if exists.
        """
        self.dsc_model = dsc_model
        self.target_dir = target_dir
        self.force_overwrite = force_overwrite
        self.progress = None

    def download(self):
        """Downloads model artifacts.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If target directory does not exist.
        """
        if os.path.exists(self.target_dir):
            if not self.force_overwrite:
                raise ValueError(
                    f"The `{self.target_dir}` directory already exists. "
                    "Set `force_overwrite` to `True` if you wish to overwrite."
                )
            shutil.rmtree(self.target_dir)
        with utils.get_progress_bar(
            ArtifactDownloader.PROGRESS_STEPS_COUNT + self.PROGRESS_STEPS_COUNT
        ) as progress:
            self.progress = progress
            self._download()
            self.progress.update(
                "Downloading model artifacts has been successfully completed."
            )
            self.progress.update("Done.")

    @abstractmethod
    def _download(self):
        """Downloads model artifacts."""


class SmallArtifactDownloader(ArtifactDownloader):
    PROGRESS_STEPS_COUNT = 3

    def _download(self):
        """Downloads model artifacts."""
        self.progress.update("Importing model artifacts from catalog")
        zip_content = self.dsc_model.get_model_artifact_content()
        with tempfile.TemporaryDirectory() as temp_dir:
            self.progress.update("Copying model artifacts to the artifact directory")
            zip_file_path = os.path.join(temp_dir, f"{str(uuid.uuid4())}.zip")
            with open(zip_file_path, "wb") as zip_file:
                zip_file.write(zip_content)
            self.progress.update("Extracting model artifacts")
            with ZipFile(zip_file_path) as zip_file:
                zip_file.extractall(self.target_dir)


class LargeArtifactDownloader(ArtifactDownloader):
    PROGRESS_STEPS_COUNT = 4

    def __init__(
        self,
        dsc_model: OCIDataScienceModel,
        target_dir: str,
        auth: Optional[Dict] = None,
        force_overwrite: Optional[bool] = False,
        region: Optional[str] = None,
        bucket_uri: Optional[str] = None,
        overwrite_existing_artifact: Optional[bool] = True,
        remove_existing_artifact: Optional[bool] = True,
    ):
        """Initializes `LargeArtifactDownloader` instance.

        Parameters
        ----------
        dsc_model: OCIDataScienceModel
            The data scince model instance.
        target_dir: str
            The target location of model after download.
        auth: (Dict, optional). Defaults to `None`.
            The default authetication is set using `ads.set_auth` API.
            If you need to override the default, use the `ads.common.auth.api_keys` or
            `ads.common.auth.resource_principal` to create appropriate authentication signer
            and kwargs required to instantiate IdentityClient object.
        force_overwrite: (bool, optional). Defaults to `False`.
            Overwrite target_dir if exists.
        region: (str, optional). Defaults to `None`.
            The destination Object Storage bucket region.
            By default the value will be extracted from the `OCI_REGION_METADATA` environment variables.
        bucket_uri: (str, optional). Defaults to None.
            The OCI Object Storage URI where model artifacts will be copied to.
            The `bucket_uri` is only necessary for uploading large artifacts which
            size is greater than 2GB. Example: `oci://<bucket_name>@<namespace>/prefix/`.
        overwrite_existing_artifact: (bool, optional). Defaults to `True`.
            Overwrite target bucket artifact if exists.
        remove_existing_artifact: (bool, optional). Defaults to `True`.
            Wether artifacts uploaded to object storage bucket need to be removed or not.
        """
        super().__init__(
            dsc_model=dsc_model, target_dir=target_dir, force_overwrite=force_overwrite
        )
        self.auth = auth or dsc_model.auth
        self.region = region or extract_region(self.auth)
        self.bucket_uri = bucket_uri
        self.overwrite_existing_artifact = overwrite_existing_artifact
        self.remove_existing_artifact = remove_existing_artifact

    def _download(self):
        """Downloads model artifacts."""
        self.progress.update(f"Importing model artifacts from catalog")

        bucket_uri = self.bucket_uri
        if not os.path.basename(bucket_uri):
            bucket_uri = os.path.join(bucket_uri, f"{self.dsc_model.id}.zip")
        elif not bucket_uri.lower().endswith(".zip"):
            bucket_uri = f"{bucket_uri}.zip"

        self.dsc_model.import_model_artifact(bucket_uri=bucket_uri, region=self.region)
        self.progress.update("Copying model artifacts to the artifact directory")
        with tempfile.TemporaryDirectory() as temp_dir:
            zip_file_path = os.path.join(temp_dir, f"{str(uuid.uuid4())}.zip")
            zip_file_path = utils.copy_file(
                uri_src=bucket_uri,
                uri_dst=zip_file_path,
                auth=self.auth,
                progressbar_description="Copying model artifacts to the artifact directory",
            )
            self.progress.update("Extracting model artifacts")
            with ZipFile(zip_file_path) as zip_file:
                zip_file.extractall(self.target_dir)

        if self.remove_existing_artifact:
            self.progress.update(
                "Removing temporary artifacts from the Object Storage bucket"
            )
            utils.remove_file(bucket_uri)
        else:
            self.progress.update()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
from typing import List, Union
from urllib.parse import urlparse

import fsspec
from ads.common.auth import default_signer
from ads.jobs.builders.base import Builder
from ads.jobs.builders.infrastructure.dataflow import DataFlow, DataFlowRun
from ads.jobs.builders.infrastructure.dsc_job import DataScienceJob, DataScienceJobRun
from ads.jobs.builders.runtimes.container_runtime import ContainerRuntime
from ads.jobs.builders.runtimes.python_runtime import (
    DataFlowRuntime,
    GitPythonRuntime,
    NotebookRuntime,
    PythonRuntime,
    Runtime,
    ScriptRuntime,
)


class Job(Builder):
    """Represents a Job defined by infrastructure and runtime.

    Examples
    --------
    Here is an example for creating and running a job::

        from ads.jobs import Job, DataScienceJob, PythonRuntime

        # Define an OCI Data Science job to run a python script
        job = (
            Job(name="<job_name>")
            .with_infrastructure(
                DataScienceJob()
                # Configure logging for getting the job run outputs.
                .with_log_group_id("<log_group_ocid>")
                # Log resource will be auto-generated if log ID is not specified.
                .with_log_id("<log_ocid>")
                # If you are in an OCI data science notebook session,
                # the following configurations are not required.
                # Configurations from the notebook session will be used as defaults.
                .with_compartment_id("<compartment_ocid>")
                .with_project_id("<project_ocid>")
                .with_subnet_id("<subnet_ocid>")
                .with_shape_name("VM.Standard.E3.Flex")
                # Shape config details are applicable only for the flexible shapes.
                .with_shape_config_details(memory_in_gbs=16, ocpus=1)
                # Minimum/Default block storage size is 50 (GB).
                .with_block_storage_size(50)
            )
            .with_runtime(
                PythonRuntime()
                # Specify the service conda environment by slug name.
                .with_service_conda("pytorch110_p38_cpu_v1")
                # The job artifact can be a single Python script, a directory or a zip file.
                .with_source("local/path/to/code_dir")
                # Environment variable
                .with_environment_variable(NAME="Welcome to OCI Data Science.")
                # Command line argument, arg1 --key arg2
                .with_argument("arg1", key="arg2")
                # Set the working directory
                # When using a directory as source, the default working dir is the parent of code_dir.
                # Working dir should be a relative path beginning from the source directory (code_dir)
                .with_working_dir("code_dir")
                # The entrypoint is applicable only to directory or zip file as source
                # The entrypoint should be a path relative to the working dir.
                # Here my_script.py is a file in the code_dir/my_package directory
                .with_entrypoint("my_package/my_script.py")
                # Add an additional Python path, relative to the working dir (code_dir/other_packages).
                .with_python_path("other_packages")
                # Copy files in "code_dir/output" to object storage after job finishes.
                .with_output("output", "oci://bucket_name@namespace/path/to/dir")
            )
        )
        # Create and Run the job
        run = job.create().run()
        # Stream the job run outputs
        run.watch()

    If you are in an OCI notebook session and you would like to use the same infrastructure
    configurations, the infrastructure configuration can be simplified.
    Here is another example of creating and running a jupyter notebook as a job::

        from ads.jobs import Job, DataScienceJob, NotebookRuntime

        # Define an OCI Data Science job to run a jupyter Python notebook
        job = (
            Job(name="<job_name>")
            .with_infrastructure(
                # The same configurations as the OCI notebook session will be used.
                DataScienceJob()
                .with_log_group_id("<log_group_ocid>")
                .with_log_id("<log_ocid>")
            )
            .with_runtime(
                NotebookRuntime()
                .with_notebook("path/to/notebook.ipynb")
                .with_service_conda(tensorflow28_p38_cpu_v1")
                # Saves the notebook with outputs to OCI object storage.
                .with_output("oci://bucket_name@namespace/path/to/dir")
            )
        ).create()
        # Run and monitor the job
        run = job.run().watch()
        # Download the notebook and outputs to local directory
        run.download(to_dir="path/to/local/dir/")


    See Also
    --------
    https://docs.oracle.com/en-us/iaas/tools/ads-sdk/latest/user_guide/jobs/index.html

    """

    _INFRASTRUCTURE_MAPPING = {
        **{item().type: item for item in [DataScienceJob]},
        "dataFlow": DataFlow,
    }
    _RUNTIME_MAPPING = {
        item().type: item
        for item in [
            PythonRuntime,
            GitPythonRuntime,
            ContainerRuntime,
            ScriptRuntime,
            NotebookRuntime,
            DataFlowRuntime,
        ]
    }

    @staticmethod
    def from_datascience_job(job_id) -> "Job":
        """Loads a data science job from OCI.

        Parameters
        ----------
        job_id : str
            OCID of an existing data science job.

        Returns
        -------
        Job
            A job instance.

        """
        dsc_infra = DataScienceJob.from_id(job_id)
        job = (
            Job(name=dsc_infra.name)
            .with_infrastructure(dsc_infra)
            .with_runtime(dsc_infra.runtime)
        )
        return job

    @staticmethod
    def datascience_job(compartment_id: str = None, **kwargs) -> List["DataScienceJob"]:
        """Lists the existing data science jobs in the compartment.

        Parameters
        ----------
        compartment_id : str
            The compartment ID for listing the jobs.
            This is optional if running in an OCI notebook session.
            The jobs in the same compartment of the notebook session will be returned.

        Returns
        -------
        list
            A list of Job objects.
        """
        return [
            Job(name=dsc_job.name)
            .with_infrastructure(dsc_job)
            .with_runtime(dsc_job.runtime)
            for dsc_job in DataScienceJob.list_jobs(compartment_id, **kwargs)
        ]

    @staticmethod
    def from_dataflow_job(job_id: str) -> "Job":
        """
        Create a Data Flow job given a job id.

        Parameters
        ----------
        job_id: str
            id of the job
        Returns
        -------
        Job
            a Job instance
        """
        df = DataFlow.from_id(job_id)
        job = Job(name=df.name).with_infrastructure(df).with_runtime(df.runtime)
        return job

    @staticmethod
    def dataflow_job(compartment_id: str = None, **kwargs) -> List["Job"]:
        """
        List data flow jobs under a given compartment.

        Parameters
        ----------
        compartment_id: str
            compartment id
        kwargs
            additional keyword arguments

        Returns
        -------
        List[Job]
            list of Job instances
        """
        return [
            Job(name=df.name).with_infrastructure(df).with_runtime(df.runtime)
            for df in DataFlow.list_jobs(compartment_id, **kwargs)
        ]

    def __init__(self, name: str = None, infrastructure=None, runtime=None) -> None:
        """Initializes a job.

        The infrastructure and runtime can be configured when initializing the job,
         or by calling with_infrastructure() and with_runtime().

        The infrastructure should be a subclass of ADS job Infrastructure, e.g., DataScienceJob, DataFlow.
        The runtime should be a subclass of ADS job Runtime, e.g., PythonRuntime, NotebookRuntime.

        Parameters
        ----------
        name : str, optional
            The name of the job, by default None.
            If it is None, a default name may be generated by the infrastructure,
            depending on the implementation of the infrastructure.
            For OCI data science job, the default name contains the job artifact name and a timestamp.
            If no artifact, a randomly generated easy to remember name with timestamp will be generated,
            like 'strange-spider-2022-08-17-23:55.02'.
        infrastructure : Infrastructure, optional
            Job infrastructure, by default None
        runtime : Runtime, optional
            Job runtime, by default None.

        """
        super().__init__()
        if name:
            self.set_spec("name", name)
        if infrastructure:
            self.with_infrastructure(infrastructure)
        if runtime:
            self.with_runtime(runtime)

    @property
    def kind(self) -> str:
        """The kind of the object as showing in YAML.

        Returns
        -------
        str
            "job"
        """
        return "job"

    @property
    def id(self) -> str:
        """The ID of the job.
        For jobs running on OCI, this is the OCID.

        Returns
        -------
        str
            ID of the job.
        """
        if self.infrastructure and hasattr(self.infrastructure, "job_id"):
            return self.infrastructure.job_id
        return None

    @property
    def name(self) -> str:
        """The name of the job.
        For jobs running on OCI, this is the display name.

        Returns
        -------
        str
            The name of the job.
        """
        return self.get_spec("name")

    @property
    def infrastructure(self) -> Union[DataScienceJob, DataFlow]:
        """The job infrastructure.

        Returns
        -------
        Infrastructure
            Job infrastructure.
        """
        return self.get_spec("infrastructure")

    @property
    def runtime(self) -> Runtime:
        """The job runtime.

        Returns
        -------
        Runtime
            The job runtime
        """
        return self.get_spec("runtime")

    def with_infrastructure(self, infrastructure) -> "Job":
        """Sets the infrastructure for the job.

        Parameters
        ----------
        infrastructure : Infrastructure
            Job infrastructure.

        Returns
        -------
        Job
            The job instance (self)
        """
        return self.set_spec("infrastructure", infrastructure)

    def with_runtime(self, runtime) -> "Job":
        """Sets the runtime for the job.

        Parameters
        ----------
        runtime : Runtime
            Job runtime.

        Returns
        -------
        Job
            The job instance (self)
        """
        return self.set_spec("runtime", runtime)

    def with_name(self, name: str) -> "Job":
        """Sets the job name.

        Parameters
        ----------
        name : str
            Job name.

        Returns
        -------
        Job
            The job instance (self)
        """
        return self.set_spec("name", name)

    def build(self) -> "Job":
        """Load default values from the environment for the job infrastructure."""
        build_method = getattr(self.infrastructure, "build", None)
        if build_method and callable(build_method):
            build_method()
        else:
            raise NotImplementedError
        return self

    def create(self, **kwargs) -> "Job":
        """Creates the job on the infrastructure.

        Returns
        -------
        Job
            The job instance (self)
        """
        infra = self.get_spec("infrastructure")
        infra.name = self.name
        self.infrastructure.create(self.runtime, **kwargs)
        self.set_spec("name", self.infrastructure.name)
        return self

    def run(
        self, name=None, args=None, env_var=None, freeform_tags=None, wait=False
    ) -> Union[DataScienceJobRun, DataFlowRun]:
        """Runs the job.

        Parameters
        ----------
        name : str, optional
            Name of the job run, by default None.
            The infrastructure handles the naming of the job run.
            For data science job, if a name is not provided,
            a default name will be generated containing the job name and the timestamp of the run.
            If no artifact, a randomly generated easy to remember name
            with timestamp will be generated, like 'strange-spider-2022-08-17-23:55.02'.
        args : str, optional
            Command line arguments for the job run, by default None.
            This will override the configurations on the job.
            If this is None, the args from the job configuration will be used.
        env_var : dict, optional
            Additional environment variables for the job run, by default None
        freeform_tags : dict, optional
            Freeform tags for the job run, by default None
        wait : bool, optional
            Indicate if this method call should wait for the job run.
            By default False, this method returns as soon as the job run is created.
            If this is set to True, this method will stream the job logs and wait until it finishes,
            similar to `job.run().watch()`.

        Returns
        -------
        Job Run Instance
            A job run instance, depending on the infrastructure.

        Examples
        --------
        To run a job and override the configurations::

            job_run = job.run(
                name="<my_job_run_name>",
                args="new_arg --new_key new_val",
                env_var={"new_env": "new_val"},
                freeform_tags={"new_tag": "new_tag_val"}
            )

        """
        return self.infrastructure.run(
            name=name,
            args=args,
            env_var=env_var,
            freeform_tags=freeform_tags,
            wait=wait,
        )

    def run_list(self, **kwargs) -> list:
        """Gets a list of runs of the job.

        Returns
        -------
        list
            A list of job run instances, the actual object type depends on the infrastructure.
        """
        return self.infrastructure.run_list(**kwargs)

    def delete(self) -> None:
        """Deletes the job from the infrastructure."""
        self.infrastructure.delete()

    def status(self) -> str:
        """Status of the job

        Returns
        -------
        str
            Status of the job
        """
        return getattr(self.infrastructure, "status", None)

    def to_dict(self) -> dict:
        """Serialize the job specifications to a dictionary.

        Returns
        -------
        dict
            A dictionary containing job specifications.
        """
        spec = {"name": self.name}
        if self.runtime:
            spec["runtime"] = self.runtime.to_dict()
        if self.infrastructure:
            spec["infrastructure"] = self.infrastructure.to_dict()
        if self.id:
            spec["id"] = self.id
        return {
            "kind": self.kind,
            # "apiVersion": self.api_version,
            "spec": spec,
        }

    @classmethod
    def from_dict(cls, config: dict) -> "Job":
        """Initializes a job from a dictionary containing the configurations.

        Parameters
        ----------
        config : dict
            A dictionary containing the infrastructure and runtime specifications.

        Returns
        -------
        Job
            A job instance

        Raises
        ------
        NotImplementedError
            If the type of the intrastructure or runtime is not supported.
        """
        if not isinstance(config, dict):
            raise ValueError("The config data for initializing the job is invalid.")
        spec = config.get("spec")

        mappings = {
            "infrastructure": cls._INFRASTRUCTURE_MAPPING,
            "runtime": cls._RUNTIME_MAPPING,
        }
        job = cls()

        for key, value in spec.items():
            if key in mappings:
                mapping = mappings[key]
                child_config = value
                if child_config.get("type") not in mapping:
                    raise NotImplementedError(
                        f"{key.title()} type: {child_config.get('type')} is not supported."
                    )
                job.set_spec(
                    key, mapping[child_config.get("type")].from_dict(child_config)
                )
            else:
                job.set_spec(key, value)

        return job

    def download(self, to_dir: str, output_uri=None, **storage_options):
        """Downloads files from remote output URI to local.

        Parameters
        ----------
        to_dir : str
            Local directory to which the files will be downloaded to.

        output_uri : (str, optional). Default is None.
            The remote URI from which the files will be downloaded.
            Defaults to None.
            If output_uri is not specified, this method will try to get the output_uri from the runtime.

        storage_options :
            Extra keyword arguments for particular storage connection.
            This method uses fsspec to download the files from remote URI.
            storage_options will to be passed into fsspec.open_files().

        Returns
        -------
        Job
            The job instance (self)

        Raises
        ------
        AttributeError
            The output_uri is not specified and the runtime is not configured with output_uri.
        """
        if not output_uri and self.runtime:
            output_uri = getattr(self.runtime, "output_uri", None)

        if not output_uri:
            raise AttributeError(
                "Please specify the output_uri or set it with a compatible runtime."
            )

        scheme = urlparse(output_uri).scheme
        if scheme == "oci" and not storage_options:
            storage_options = default_signer()
        fs = fsspec.filesystem(scheme, **storage_options)
        fs.get(self.runtime.output_uri, to_dir, recursive=True)
        return self



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import ast
import base64
import html
import io
import math
import os
import warnings
import re
from collections import defaultdict
import inspect
import importlib
from typing import List, Union
import fsspec

# from pandas.io.common import _compression_to_extension

from numbers import Number
from urllib.parse import urlparse

import numpy as np
import pandas as pd

from pandas.core.dtypes.common import (
    is_numeric_dtype,
    is_bool_dtype,
    is_categorical_dtype,
    is_datetime64_any_dtype,
    is_float_dtype,
)

from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common import utils
from ads.dataset import logger
from ads.type_discovery.type_discovery_driver import TypeDiscoveryDriver
from ads.type_discovery.typed_feature import (
    ContinuousTypedFeature,
    DateTimeTypedFeature,
    CategoricalTypedFeature,
    UnknownTypedFeature,
    OrdinalTypedFeature,
    DocumentTypedFeature,
)


class DatasetDefaults:
    sampling_confidence_level = 95
    sampling_confidence_interval = 1.0


_known_db_protocols = {"sqlite", "ADB", "oracle+cx_oracle"}


def concatenate(X, y):
    if isinstance(X, pd.DataFrame) or isinstance(X, pd.Series):
        return pd.concat([X, y], axis=1)
    else:
        return X.assign(**{y.name: y})


def fix_column_names(X):
    X.columns = X.columns.astype("str").str.strip().str.replace(" ", "_")
    return X


def convert_columns(df, feature_metadata=None, dtypes=None):
    if feature_metadata is not None:
        dtypes = {}
        for feature in feature_metadata:
            dtype = get_dtype(feature_metadata[feature], df[feature].dtype)
            if dtype is not None:
                dtypes[feature] = dtype
    return df.astype(dtypes)


def get_dtype(feature_type, dtype):
    if isinstance(feature_type, ContinuousTypedFeature) or isinstance(
        feature_type, OrdinalTypedFeature
    ):
        return dtype.name if is_numeric_dtype(dtype) else "float"
    elif isinstance(feature_type, DateTimeTypedFeature):
        return "datetime64[ns]" if not dtype.name.startswith("datetime") else dtype
    elif isinstance(feature_type, CategoricalTypedFeature):
        return "bool" if is_bool_dtype(dtype) else "category"


def get_feature_type(name, series):
    if is_bool_dtype(series) or is_categorical_dtype(series):
        return CategoricalTypedFeature.build(name, series)
    elif is_numeric_dtype(series):
        if is_float_dtype(series):
            return ContinuousTypedFeature.build(name, series)
        else:
            return OrdinalTypedFeature.build(name, series)
    elif is_datetime64_any_dtype(series):
        return DateTimeTypedFeature.build(name, series)
    else:
        return UnknownTypedFeature.build(name, series)


def convert_to_html(plot):
    img = io.BytesIO()
    plot.savefig(img, format="png", bbox_inches="tight")
    img.seek(0)
    encoded = base64.b64encode(img.getvalue())
    return '<img width=95%" src="data:image/png;base64, {}"><hr><br>'.format(
        encoded.decode("utf-8")
    )


def _num_partitions_for_dataframe(df):
    # takes pandas dataframe, guesses good number of partitions
    return utils.get_cpu_count() if df.shape[0] > 1000 * utils.get_cpu_count() else 1


class ElaboratedPath:
    """
    The Elaborated Path class unifies all of the operations and information related to a path or pathlist.
    Whether the user wants to
    An Elaborated path can accept any of the following as a valid source:
    * A single path
    * A glob pattern path
    * A directory
    * A list of paths (Note: all of these paths must be from the same filesystem AND have the same format)
    * A sqlalchemy connection url
    """

    def __init__(
        self,
        source: Union[str, List[str]],
        format: str = None,
        name: str = None,
        **kwargs,
    ):
        """
        :param source:
        :param format:
        :param kwargs:

        By the end of this method, this class needs to have paths, format, and name ready
        """
        self._kwargs = kwargs
        self._format = format
        self._name = name
        if isinstance(source, str):
            self._original_source = source
            self._determine_protocol_type()
            if self._type == "db":
                self._paths = [self._original_source]
            else:
                self._elaborate_path()
        elif isinstance(source, list) and all(isinstance(file, str) for file in source):
            assert len(source) > 0, "Error, the source you passed in was an empty list."
            self._original_source = source[0]
            self._paths = source
            self._type = "list"
        else:
            raise ValueError(f"Source argument not understood: {source}")
        if self.num_paths == 0:
            raise FileNotFoundError(
                f"Error: We could not find any files associated with the source: "
                f"{source}. Double check that this source is a valid glob pattern,"
                f" directory, or path."
            )
        self._determine_format()
        self._determine_name()

    @property
    def paths(self) -> List[str]:
        """
        :return: a list of str
            Each element will be a valid path
        """
        return self._paths

    @property
    def num_paths(self) -> int:
        """
        This method will return the number of paths found with the associated original glob, folder, or path.
        If this returns 0,
        :return:
        """
        return len(self._paths)

    @property
    def name(self) -> str:
        return self._name

    @property
    def format(self) -> str:
        return self._format

    def _determine_name(self):
        if self._name is None:
            if self._type == "list":
                self._name = (
                    f"DataFrame from [{os.path.basename(self._original_source)}, ...]"
                )
            elif self._type == "glob":
                self._name = f"DataFrame from {os.path.basename(self._original_source)}"
            else:
                self._name = f"DataFrame from {urlparse(self._original_source).scheme}"

    def _determine_format(self):
        """
        Infer format from the path.

        If its  a compressed file, returns the extension before compression extension.
        If the extension cannot be inferred,  returns None

        Parameters
        ----------
        path : ElaboratedPath
            an ElaboratedPath object

        Returns
        -------
        format : str
        """
        if self._format in [None, "infer"]:
            format_keys = []
            for i in range(min(self.num_paths, 5)):
                format_keys.append(self._remove_compressions(self.paths[i]))
            if len(format_keys) == 0:
                raise ValueError(
                    f"Could not determine the format key for source: {self._original_source}"
                )
            if format_keys.count(format_keys[0]) != len(format_keys):
                raise ValueError(
                    f"Got multiple formats from the source: {self._original_source}. Run again "
                    f'using the format parameter. Ex: format=<your format key, like: "csv", "hdf", etc.>'
                )
            self._format = format_keys[0]
        else:
            self._format = self._format.lower()

    def _elaborate_path(self):
        self._paths = self._fs.glob(self._original_source)
        if self._protocol != "":
            self._paths = [f"{self._protocol}://{p}" for p in self._paths]

    def _determine_protocol_type(self):
        self._protocol = urlparse(self._original_source).scheme

        if self._kwargs.get("fs") is not None:
            self._fs = self._kwargs.pop("fs")
            self._type = "glob"
        elif self._original_source.startswith("oracle+cx_oracle://"):
            self._protocol = "oracle+cx_oracle"
            self._type = "db"
        else:
            try:
                self._fs = fsspec.filesystem(
                    self._protocol, **self._kwargs.get("storage_options", dict())
                )
                self._type = "glob"
            except ValueError:
                try:
                    self.engine = utils.get_sqlalchemy_engine(
                        self._original_source, **self._kwargs
                    )
                    self._type = "db"
                except:
                    if self._protocol in _known_db_protocols:
                        self._type = "db"
                    else:
                        raise ValueError(
                            f"Error in trying to understand the protocol for source: "
                            f"{self._original_source}. The protocol found: {self._protocol} is not "
                            f"registered with fsspec or sqlalchemy"
                        )

    @staticmethod
    def _remove_compressions(filename: str):
        _compression_to_extension = [
            ".gz",
            ".bz2",
            ".zip",
            ".xz",
            ".zst",
            ".tar",
            ".tar.gz",
            ".tar.xz",
            ".tar.bz2",
        ]
        for compression in _compression_to_extension:
            if filename.strip().endswith(compression):
                return ElaboratedPath._remove_compressions(
                    os.path.splitext(filename.rstrip("/*"))[0]
                )
        format = os.path.splitext(filename.rstrip("/*"))[1][1:].lower()
        return format.lower() if format != "" else None


class DatasetLoadException(BaseException):
    def __init__(self, exc_msg):
        self.exc_msg = exc_msg

    def __str__(self):
        return self.exc_msg


def _get_dtype_from_error(e):
    error_string = str(e)

    if "mismatched dtypes" in error_string.lower():

        # For the mismatched dtypes error, dask either returns a error message containing the dtype argument
        # to specify, or  the found and expected dtypes in a table format, depending on what stage
        # the type inferencing fails. The below logic supports building the dtype dictionary for both cases
        found_dtype_dict_str_list = re.findall(
            r"dtype=({[^{}]+})", error_string, re.MULTILINE
        )
        if found_dtype_dict_str_list:
            found_dtype_dict = ast.literal_eval(found_dtype_dict_str_list[0])
        else:
            found_dtype_dict = _find_dtypes_from_table(error_string)
        if found_dtype_dict:
            logger.warning(
                "Dask type-inference/coercion failed. Retrying with "
                f"dtype={found_dtype_dict}.",
                exc_info=True,
            )
            return found_dtype_dict
    return None


def _find_dtypes_from_table(error_string):
    error_lines = error_string.splitlines()
    dtypes = {}
    # matches '| Column | Found | Expected |'
    pattern = re.compile(
        "\\s*\\|\\s*Column\\s*\\|\\s*Found\\s*\\|\\s*Expected\\s*\\|\\s*"
    )
    for i, line in enumerate(error_lines):
        if re.match(pattern, line):
            for j in range(i + 2, len(error_lines)):
                # extracts column_name and found_dtype from  '| <column_name> | <found_dtype> | <expected_dtype |'
                dtype_suggestion = re.compile("\\s*\\|([^\\|]+)\\|([^\\|]+)\\|.*")
                match_groups = re.match(dtype_suggestion, error_lines[j])
                if match_groups is None:
                    break
                dtypes[match_groups.group(1).strip()] = match_groups.group(2).strip()
    return dtypes


def rename_duplicate_cols(original_cols):
    seen_col_names = defaultdict(int)
    new_cols = []
    for col in original_cols:
        # remove any spaces form column names
        if isinstance(col, str):
            col.replace(" ", "_")
        if col not in seen_col_names:
            new_cols.append(col)
        else:
            dup_count = seen_col_names[col]
            new_cols.append(f"{col}.{dup_count}")
        seen_col_names[col] += 1
    assert len(new_cols) == len(
        original_cols
    ), "There has been an error in re-naming duplicate columns"
    return new_cols


def write_parquet(
    path,
    data,
    engine="fastparquet",
    metadata_dict=None,
    compression=None,
    storage_options=None,
):
    """
    Uses fast parquet to write dask dataframe and custom metadata in parquet format

    Parameters
    ----------
    path : str
        Path to write to
    data : pandas.DataFrame
    engine : string
        "auto" by default
    metadata_dict : Deprecated, will not pass through
    compression : {{'snappy', 'gzip', 'brotli', None}}, default 'snappy'
        Name of the compression to use
    storage_options : dict, optional
        storage arguments required to read the path

    Returns
    -------
    str : the file path the parquet was written to
    """
    assert isinstance(data, pd.DataFrame)
    if metadata_dict is not None:
        warnings.warn(
            "The `metadata_dict` argument is deprecated and has no effect on this method.",
            DeprecationWarning,
            stacklevel=2,
        )
    data.to_parquet(
        path,
        engine=engine,
        compression=compression,
        storage_options=storage_options,
    )
    return path


def is_text_data(df, target=None):
    if len(df.columns.values) == 2:
        feature_name = (
            list(set(df.columns.values) ^ set([target]))[0]
            if target
            else list(set(df.columns.values))[0]
        )
    elif len(df.columns.values == 1):
        feature_name = df.columns.values[0]
    else:
        return False
    return isinstance(
        TypeDiscoveryDriver().discover(feature_name, df[feature_name]),
        DocumentTypedFeature,
    )


def generate_sample(
    df: pd.DataFrame,
    n: int,
    confidence_level: int = DatasetDefaults.sampling_confidence_level,
    confidence_interval: float = DatasetDefaults.sampling_confidence_interval,
    **kwargs,
):
    min_size_to_sample = min(n, 10000)

    sample_size = None

    if "sample_max_rows" in kwargs:
        requested_sample_size = int(kwargs["sample_max_rows"])

        if requested_sample_size < 0:
            sample_size = calculate_sample_size(
                n, min_size_to_sample, confidence_level, confidence_interval
            )
        else:
            if min_size_to_sample < requested_sample_size < n:
                logger.info(
                    f"Downsampling from {n} rows, to the user specified {requested_sample_size} rows for graphing."
                )
                sample_size = requested_sample_size
            elif requested_sample_size >= n:
                logger.info(f"Using the entire dataset of {n} rows for graphing.")
                sample_size = n
            else:
                sample_size = min_size_to_sample
                logger.info(
                    f"Downsampling from {n} rows, to {sample_size} rows for graphing."
                )

    if sample_size and len(df) > sample_size:
        frac = min(1.0, sample_size * 1.05 / n)
        df = df.sample(frac=frac, random_state=42)
        return df.head(sample_size) if len(df) > sample_size else df
    else:
        return df


def calculate_sample_size(
    population_size, min_size_to_sample, confidence_level=95, confidence_interval=1.0
):
    """Find sample size for a population using Cochran’s Sample Size Formula.
     With default values for confidence_level (percentage, default: 95%)
     and confidence_interval (margin of error, percentage, default: 1%)

    SUPPORTED CONFIDENCE LEVELS: 50%, 68%, 90%, 95%, and 99% *ONLY* - this
    is because the Z-score is table based, and I'm only providing Z
    for common confidence levels.
    """

    if population_size < min_size_to_sample:
        return None

    confidence_level_constant = {
        50: 0.67,
        68: 0.99,
        90: 1.64,
        95: 1.96,
        99: 2.57,
        99.5: 2.807,
        99.9: 3.291,
    }

    p = 0.5
    e = confidence_interval / 100.0
    N = population_size
    n_0 = 0.0
    n = 0.0

    Z = confidence_level_constant.get(confidence_level, 99)

    n_0 = ((Z**2) * p * (1 - p)) / (e**2)
    n = n_0 / (1 + ((n_0 - 1) / float(N)))

    sample_size = max(int(math.ceil(n)), min_size_to_sample)

    logger.info(f"Downsampling from {population_size} rows to {sample_size} rows.")

    return sample_size


def map_types(types):
    for column in types:
        if types[column] == "continuous":
            types[column] = "float64"
        elif types[column] == "ordinal":
            types[column] = "int64"
        elif types[column] == "categorical":
            types[column] = "category"
        elif types[column] == "datetime":
            types[column] = "datetime64[ns]"
    return types


@runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
@runtime_dependency(module="graphviz", install_from=OptionalDependency.VIZ)
def visualize_transformation(transformer_pipeline, text=None):
    dot = graphviz.Digraph()

    # show a single node for paritions
    dot.attr(
        "node",
        shape="tab",
        style="filled",
        fontname="courier",
        fontsize="12",
        fontcolor="white",
        resolution="144",
    )
    if text:
        dot.node("partitions", text, margin="0.25", fillcolor="dimgray")

    dot.attr(
        "node",
        shape="component",
        style="filled",
        fontname="courier",
        fontsize="10",
        fontcolor="black",
        resolution="144",
    )
    for step in transformer_pipeline.steps:
        name, clazz, clazzname, is_ads = (
            step[0],
            step[1],
            step[1].__class__.__name__,
            "ads" in str(step[1].__class__),
        )
        ads_node = str(step[1].__class__.__name__) in [
            "AutoMLPreprocessingTransformer",
            "DataFrameTransformer",
            "RecommendationTransformer",
            "AutoMLFeatureSelection",
            "FeatureEngineeringTransformer",
        ]
        if ads_node:
            text2html = "< {} >".format(
                html.escape(step[1].__repr__()).replace("\n", "<br/>")
            )
            dot.node(name, text2html, margin="0.25", fillcolor="gold2")
        else:
            dot.node(name, name.rsplit("/")[0], fillcolor="azure")

    def format_label(stage):
        if "FunctionTransformer" in str(transformer_pipeline.steps[stage][1].__class__):
            return "<<font face='courier' point-size='10'>&nbsp;<b>{}</b>&nbsp;</font>>".format(
                html.escape(transformer_pipeline.steps[stage][1].func.__name__)
            )
        else:
            is_ads = "ads" in str(transformer_pipeline.steps[stage][1].__class__)
            return "<<font face='courier' point-size='10'>&nbsp;<b>{}</b>&nbsp;</font>>".format(
                transformer_pipeline.steps[stage][1].__class__.__name__
            )

    edges = [x[0] for x in transformer_pipeline.steps]
    for i, edge in enumerate(list(zip(edges[:-1], edges[1:]))):
        dot.edge(*edge, len="1.00", label=format_label(i))

    # terminus node
    dot.node("terminus", "", shape="terminator", fillcolor="white")
    dot.edge(edges[-1], "terminus", len="1.00", label=format_label(len(edges) - 1))

    graph = graphviz.Source(dot)

    from IPython.core.display import display, SVG

    display(SVG(graph.pipe(format="svg")))


def up_sample(df, target, sampler="default", feature_types=None):
    """
    Fixes imbalanced dataset by up-sampling

    Parameters
    ----------
    df : Union[pandas.DataFrame, dask.dataframe.core.DataFrame]
    target : name of the target column in df
    sampler: Should implement fit_resample(X,y) method
    fillna: a dictionary contains the column name as well as the fill value,
            only needed when the column has missing values

    Returns
    -------
    upsampled_df : Union[pandas.DataFrame, dask.dataframe.core.DataFrame]
    """
    if sampler != "default":
        if inspect.getattr_static(sampler, "fit_resample", None) is None:
            raise AttributeError("`sampler` object must has method `fit_resample`.")
        else:
            # exactly two input args X, y will be passed to fit_resample()
            # check signature of fit_sample
            num_no_default_params = 0
            sig = inspect.signature(sampler.fit_resample)
            for param in sig.parameters.values():
                if param.default is param.empty:
                    num_no_default_params += 1
            if len(sig.parameters) < 2 or num_no_default_params > 2:
                raise RuntimeError(
                    "The signature for `sampler.fit_resample` has to be `fit_resample(X, y)`."
                )

    X = df.drop(target, axis=1)
    y = df[target]

    feature_types = feature_types if feature_types is not None else {}

    columns_with_nans = X.columns.values[X.isna().any()]
    if len(columns_with_nans) > 0:
        fill_nan_dict = {}
        for column in columns_with_nans:
            if column in feature_types and "mode" in feature_types[column]["stats"]:
                fill_nan_dict[column] = feature_types[column]["stats"]["mode"]
            elif column in feature_types and "mean" in feature_types[column]["stats"]:
                fill_nan_dict[column] = feature_types[column]["stats"]["mean"]
            elif column in feature_types and "median" in feature_types[column]["stats"]:
                fill_nan_dict[column] = feature_types[column]["stats"]["median"]
            else:
                logger.warning(
                    "Sampling from a column that has missing values may cause an error."
                )
        X = X.fillna(fill_nan_dict)

    if sampler == "default":
        imblearn_found = importlib.util.find_spec("imblearn") is not None
        if not imblearn_found:
            raise ModuleNotFoundError(
                """
                Required package for up-sampling `imblearn` not found.
                Install `imblearn` with `pip install imbalanced-learn`
                and rerun to enable up-sampling.
            """
            )
        else:
            sampler = _get_imblearn_sampler(X, y)
    return _sample(sampler, X, y)


def _get_imblearn_sampler(X, y):
    from imblearn.over_sampling import SMOTE, RandomOverSampler

    categorical_feature_indices = [
        X.columns.get_loc(c)
        for c in X.select_dtypes(
            include=["category", "object", "datetime64"]
        ).columns.values
    ]

    if len(categorical_feature_indices) > 0:
        logger.info(
            """
            Using the default `RandomOverSampler` sampler. Use `sample` to specify a sampler.
            Classes will be equalized.
            You can also pass in other samplers such as `imblearn.SMOTENC` instead, e.g.
            sampler = SMOTENC(categorical_features=categorical_feature_indices)
            ds.up_sample(sampler=sampler)
        """
        )
        return RandomOverSampler(random_state=42)

    min_sample_size = y.value_counts().min()

    k_neighbors = min(min_sample_size - 1, 5)
    if k_neighbors == 0:
        logger.warning(
            f"""k_neighbors is 0 as in the target there exists a class label that appeared only once.
                SMOTE will fail. Default to RandomOverSampler.
            """
        )
        return RandomOverSampler(random_state=42)
    else:
        if 5 > k_neighbors > 0:
            logger.info(
                f"`k_neighbors()` of SMOTE has changed to {k_neighbors}"
                " as the target has at least one class which appeared "
                f"only {min_sample_size} times in the data. "
            )
        logger.info("Using SMOTE for over sampling. Classes will be equalized.")
        return SMOTE(random_state=42, k_neighbors=k_neighbors)


def down_sample(df, target):
    """
    Fixes imbalanced dataset by down-sampling

    Parameters
    ----------
    df : pandas.DataFrame
    target : name of the target column in df

    Returns
    -------
    downsampled_df : pandas.DataFrame
    """
    dfs = []
    target_value_counts = df[target].value_counts()
    min_key = min(target_value_counts.iteritems(), key=lambda k: k[1])
    for key, value in target_value_counts.iteritems():
        if key != min_key[0]:
            dfs.append(
                df[df[target] == key].sample(frac=1 - ((value - min_key[1]) / value))
            )
    dfs.append(df[df[target] == min_key[0]])
    return pd.concat(dfs)


def _sample(sampler, X, y):
    if isinstance(y, pd.Series) and (
        isinstance(y[0], bool) or isinstance(y[0], np.bool_)
    ):
        y_trans = y.astype(int)  ## Convert to ints to let SMOTE sample properly
        X_resampled, y_resampled = sampler.fit_resample(X=X, y=y_trans)
    else:
        X_resampled, y_resampled = sampler.fit_resample(X=X, y=y)

    if not isinstance(X_resampled, pd.DataFrame):
        X_resampled = pd.DataFrame(X_resampled, columns=X.columns.values)
    if not isinstance(y_resampled, pd.Series):
        y_resampled = pd.DataFrame(y_resampled, columns=[y.name])[y.name]

    for k in X.dtypes.keys():
        X_resampled[k] = X_resampled[k].astype(X.dtypes[k].name)
    balanced_df = concatenate(X_resampled, y_resampled)
    return balanced_df


def get_fill_val(feature_types, column, action, constant="constant"):
    # action can be one of the following
    # "Fill missing values with mean", "Fill missing values with median",
    # "Fill missing values with frequent", "Fill missing values with constant"
    action_ = action.split(" ")[-1]
    fill_type = "mode" if action_ == "frequent" else action_
    try:
        fill_val = (
            feature_types[column].meta_data["stats"][fill_type]
            if action_ != "constant"
            else constant
        )
        fill_val = round(fill_val, 4) if isinstance(fill_val, Number) else fill_val
    except:
        fill_val = None
    return fill_val


def parse_apache_log_str(x):
    """
    Returns the string delimited by two characters.

    Source: https://mmas.github.io/read-apache-access-log-pandas
    Example:
        `>>> parse_str('[my string]')`
        `'my string'`
    """
    if x is not None:
        return x[1:-1]
    return np.nan


def parse_apache_log_datetime(x):
    """
    Parses datetime with timezone formatted as:
        `[day/month/year:hour:minute:second zone]`

    Source: https://mmas.github.io/read-apache-access-log-pandas
    Example:
        `>>> parse_datetime('13/Nov/2015:11:45:42 +0000')`
        `datetime.datetime(2015, 11, 3, 11, 45, 4, tzinfo=<UTC>)`

    Due to problems parsing the timezone (`%z`) with `datetime.strptime`, the
    timezone will be obtained using the `pytz` library.
    """
    import pytz
    from datetime import datetime

    dt = datetime.strptime(x[1:-7], "%d/%b/%Y:%H:%M:%S")
    dt_tz = int(x[-6:-3]) * 60 + int(x[-3:-1])
    return dt.replace(tzinfo=pytz.FixedOffset(dt_tz))


def deprecate_variable(old_var, new_var, warning_msg, warning_type):
    if old_var is not None:
        warnings.warn(warning_msg, warning_type)
        return old_var
    return new_var


def deprecate_default_value(var, old_value, new_value, warning_msg, warning_type):
    if var == old_value:
        warnings.warn(warning_msg, warning_type)
        return new_value
    else:
        return var


def _log_yscale_not_set():
    logger.info(
        "`yscale` parameter is not set. Valid values are `'linear'`, `'log'`, `'symlog'`."
    )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from ads.dataset.dataset_with_target import ADSDatasetWithTarget


class RegressionDataset(ADSDatasetWithTarget):
    def __init__(self, df, sampled_df, target, target_type, shape, **kwargs):
        ADSDatasetWithTarget.__init__(
            self, df, sampled_df, target, target_type, shape, **kwargs
        )
